import json
from types import SimpleNamespace
from typing import Annotated, Sequence
from unittest import mock
from unittest.mock import patch

import httpx
import pytest
from pydantic import BaseModel, BeforeValidator, PlainSerializer, WithJsonSchema, computed_field

import fbnconfig.resource_abc
from fbnconfig import create_client
from fbnconfig.deploy import Action, Deployment, dump_deployment, run, undump_deployment
from fbnconfig.resource_abc import CamelAlias, Resource, get_resource_class, register_resource

TEST_BASE = "https://foo.lusid.com"

log_entity = "deployment"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWithNoRemoteResources:
    client = create_client(TEST_BASE, "xyz")

    def test_handles_429_on_get(self, respx_mock):
        # given the server returns a 429 before returning the real response
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            side_effect=[
                httpx.Response(429, json={}, headers={"retry-after": "0"}),
                httpx.Response(200, json={"values": []}),
            ]
        )
        # when we run an empty deployment
        deployment = Deployment("whatever", [])
        run(self.client, deployment)
        # then the 429 gets retried and we continue
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert request.url.params._dict == {"filter": ["fields[deployment] eq 'whatever'"]}
        # but no other http requests are made

    def test_run_with_empty_deploy_does_nothing(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        # when we run an empty deployment
        deployment = Deployment("whatever", [])
        actions = run(self.client, deployment)
        # then the log is searched
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert request.url.params._dict == {"filter": ["fields[deployment] eq 'whatever'"]}
        # but no other http requests are made
        # and an empty actions list is returned
        assert actions == []

    def test_run_with_one_resource_creates_it(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={})
        )
        # when we run a deployment with one resource
        resource = mock.Mock(spec=Resource)
        resource.deps = mock.Mock(return_value=[])
        resource.id = "23"
        resource.create = mock.Mock(return_value={"some_variable": "x"})
        deployment = Deployment("whatever", [resource])
        actions = run(self.client, deployment)
        # then the create method on the resource is called
        resource.create.assert_called_once_with(self.client)
        # and the create is recorded in the log
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert json.loads(request.content) == {
            "description": "...",
            "displayName": "23",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "Mock"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "23"},
                {"name": "state", "value": '{"some_variable": "x"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": log_entity,
                    "identifierType": "resource",
                    "identifierValue": "whatever_23",
                }
            ],
        }
        # and the create is in the actions
        assert actions == [Action(id="23", type="Mock", change="create")]

    def test_run_with_one_resource_empty_state(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={})
        )
        # when we run a deployment with one resource
        resource = mock.Mock(spec=Resource)
        resource.deps = mock.Mock(return_value=[])
        resource.id = "23"
        # and the new state value is an empty dict
        resource.create = mock.Mock(return_value={})
        deployment = Deployment("whatever", [resource])
        actions = run(self.client, deployment)
        # then the create method on the resource is called
        resource.create.assert_called_once_with(self.client)
        # and the create is recorded in the log
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert json.loads(request.content) == {
            "description": "...",
            "displayName": "23",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "Mock"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "23"},
                {"name": "state", "value": "{}"},
            ],
            "identifiers": [
                {
                    "identifierScope": log_entity,
                    "identifierType": "resource",
                    "identifierValue": "whatever_23",
                }
            ],
        }
        # and the action is to create
        assert actions == [Action(id="23", type="Mock", change="create")]

    def test_run_with_two_resources(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={})
        )
        # and desired resources one and two are independent
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[])
        resource2.id = "two"
        resource2.create = mock.Mock(return_value={"some_variable": "y"})
        # when we run a deployment with one and two
        deployment = Deployment("whatever", [resource1, resource2])
        actions = run(self.client, deployment)
        # then the create methods are called on both
        resource1.create.assert_called_once_with(self.client)
        resource2.create.assert_called_once_with(self.client)
        # and the create is recorded in the log for both in the order they were
        # added to the deployment
        request = respx_mock.calls[-2].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert json.loads(request.content)["displayName"] == "one"
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert json.loads(request.content)["displayName"] == "two"
        # and they appear in order in the actions
        assert actions == [
            Action(id="one", type="Mock", change="create"),
            Action(id="two", type="Mock", change="create"),
        ]

    def test_run_logs_unique_dependencies(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={})
        )
        # and resources "one" and "two" where "two" depends on "one"
        # twice
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1, resource1])  # double dependency
        resource2.id = "two"
        resource2.create = mock.Mock(return_value={"some_variable": "x"})
        # when we deploy
        deployment = Deployment("whatever", [resource1, resource2])
        run(self.client, deployment)
        # then the create method on each of them is called
        resource1.create.assert_called_once_with(self.client)
        resource2.create.assert_called_once_with(self.client)
        # and the create is recorded in the log for both in dependency order
        request = respx_mock.calls[-2].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert json.loads(request.content)["displayName"] == "one"
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        two_log = json.loads(request.content)
        assert two_log["displayName"] == "two"
        # and resource two has uniquely recorded dependency on resource one
        assert next(
            field["value"] for field in two_log["fields"] if field["name"] == "dependencies"
        ) == ["one"]

    def test_run_with_one_resource_and_dependency(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={})
        )
        # and resources "one" and "two" where "two" depends on "one"
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1])
        resource2.id = "two"
        resource2.create = mock.Mock(return_value={"some_variable": "x"})
        # when we deploy "two" only
        deployment = Deployment("whatever", [resource2])
        actions = run(self.client, deployment)
        # then the create method on each of them is called
        resource1.create.assert_called_once_with(self.client)
        resource2.create.assert_called_once_with(self.client)
        # and the create is recorded in the log for both in dependency order
        request = respx_mock.calls[-2].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert json.loads(request.content)["displayName"] == "one"
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        two_log = json.loads(request.content)
        assert two_log["displayName"] == "two"
        # and resource two has recorded a dependency on resource one
        assert next(f["value"] for f in two_log["fields"] if f["name"] == "dependencies") == ["one"]
        # and they appear in order in the actions
        assert actions == [
            Action(id="one", type="Mock", change="create"),
            Action(id="two", type="Mock", change="create"),
        ]

    def test_run_with_deuplicate_resourcei_ids(self, respx_mock):
        # given two resources using the same resource id and a dependency between them
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "duplicate-id"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1])
        resource2.id = "duplicate-id"
        resource2.create = mock.Mock(return_value={"some_variable": "x"})
        # when we deploy resource2
        deployment = Deployment("whatever", [resource2])
        # then an error is raised because resource2 depends on resource1
        # but they have the same resource id
        with pytest.raises(RuntimeError) as ex:
            run(self.client, deployment)
        assert "duplicate-id has been used on more than one resource" in str(ex.value)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWithExistingResources:
    client = httpx.Client(base_url=TEST_BASE)

    @pytest.fixture
    def existing_resource(self):
        return {
            "description": "...",
            "displayName": "one",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "FileResource"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "one"},
                {"name": "state", "value": '{"some_variable": "x"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": log_entity,
                    "identifierType": "resource",
                    "identifierValue": "whatever_one",
                }
            ],
            "version": {"asAtModified": "2024-02-20T15:39:58.1008350+00:00"},
        }

    @pytest.fixture
    def dependent_resource(self):
        # this resource "two" depends on resource "one"
        return {
            "description": "...",
            "displayName": "two",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "FileResource"},
                {"name": "dependencies", "value": ["one"]},
                {"name": "resource", "value": "two"},
                {"name": "state", "value": '{"other_variable": "x"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": log_entity,
                    "identifierType": "resource",
                    "identifierValue": "whatever_two",
                }
            ],
            "version": {"asAtModified": "2024-02-20T15:39:58.1008350+00:00"},
        }

    def test_run_existing_resource_calls_update_no_change(self, respx_mock, existing_resource):
        # given that resource "one" already exists in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource]})
        )
        # and the desired resource one
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.__name__ = "FileResource"
        # and it is the same as the remote
        resource1.update = mock.Mock(return_value=None)
        # when we deploy "one"
        deployment = Deployment("whatever", [resource1])
        actions = run(self.client, deployment)
        # then the update method is called with the log state
        resource1.update.assert_called_once_with(self.client, SimpleNamespace(some_variable="x"))
        # and the log is not updated
        # but nochange is in the actions
        assert actions == [Action(id="one", type="Mock", change="nochange")]

    def test_run_existing_resource_calls_update_new_state(self, respx_mock, existing_resource):
        # given that resource "one" already exists in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource]})
        )
        respx_mock.post(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={})
        )
        # and a desired resource "one"
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.__name__ = "FileResource"
        # and it is different to the remote state (it returns not None)
        resource1.update = mock.Mock(return_value={"some_variable": "abcdefg"})
        # when we deploy "one"
        deployment = Deployment("whatever", [resource1])
        actions = run(self.client, deployment)
        # then the update method is called with the remote log state
        resource1.update.assert_called_once_with(self.client, SimpleNamespace(some_variable="x"))
        # and the state returned by update is recorded in the log
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        content = json.loads(request.content)
        assert content["displayName"] == "one"
        new_state = next(field["value"] for field in content["fields"] if field["name"] == "state")
        assert json.loads(new_state) == {"some_variable": "abcdefg"}
        # and the action is update
        assert actions == [Action(id="one", type="Mock", change="update")]

    def test_run_existing_resource_gets_deleted_when_not_desired(self, respx_mock, existing_resource):
        # given that resource "one" already exists in the log
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource]})
        )
        respx_mock.delete(
            f"/api/api/customentities/~{log_entity}/resource/whatever_one?identifierScope={log_entity}"
        ).mock(return_value=httpx.Response(200, json={}))
        # and "one" is not in the deployment
        deployment = Deployment("whatever", [])
        from fbnconfig import drive

        mock_resource = mock.Mock(spec=drive.FileResource)
        mock_resource.delete = mock.Mock(return_value=None)
        with mock.patch(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"FileResource": {"class": mock_resource}}
        ):
            actions = run(self.client, deployment)
            # then the FileResource.delete function is called (the patched one for this test)
            mock_resource.delete.assert_called_once_with(self.client, SimpleNamespace(some_variable="x"))
            # and the log record is deleted
            request = respx_mock.calls[-1].request
            assert request.method == "DELETE"
            # and the url matches the one mocked above
            # and the action is update
            assert actions == [Action(id="one", type="FileResource", change="remove")]

    def test_run_implicit_resource_does_not_get_deleted(
        self, respx_mock, existing_resource, dependent_resource
    ):
        # given that resource "one" and "two already exists in the log and "two" depends on "one"
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource, dependent_resource]})
        )
        # and desired resources "one" and "two" match what has been deployed
        #    where "two" depends on "one"
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.update = mock.Mock(return_value=None)
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1])  # two depends on one
        resource2.id = "two"
        resource2.update = mock.Mock(return_value=None)
        # and "two" is explicitly in the deployment but "one" is only implicity
        # because two depends on "one"
        deployment = Deployment("whatever", [resource2])
        from fbnconfig import drive

        mock_resource = mock.Mock(spec=drive.FileResource)
        mock_resource.delete = mock.Mock(return_value=None)
        with mock.patch(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"FileResource": {"class": mock_resource}}
        ):
            run(self.client, deployment)
            # then no delete calls are made because we should keep both resources
            assert mock_resource.delete.mock_calls == []

    def test_run_dependent_resources_get_deleted_in_order(
        self, respx_mock, existing_resource, dependent_resource
    ):
        # given resources "one" and "two" where "two"depends on "one"
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": [dependent_resource, existing_resource]})
        )
        respx_mock.delete(
            f"/api/api/customentities/~{log_entity}/resource/whatever_one?identifierScope={log_entity}",
            name="delete_one",
        ).mock(return_value=httpx.Response(200, json={}))
        respx_mock.delete(
            f"/api/api/customentities/~{log_entity}/resource/whatever_two?identifierScope={log_entity}",
            name="delete_two",
        ).mock(return_value=httpx.Response(200, json={}))
        # and the deployment is empty
        deployment = Deployment("whatever", [])
        from fbnconfig import drive

        mock_resource = mock.Mock(spec=drive.FileResource)
        mock_resource.delete = mock.Mock(return_value=None)
        with mock.patch(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"FileResource": {"class": mock_resource}}
        ):
            run(self.client, deployment)
            # then the FileResource.delete function is called for both, dependent resource first
            assert mock_resource.delete.mock_calls == [
                mock.call(self.client, SimpleNamespace(other_variable="x")),
                mock.call(self.client, SimpleNamespace(some_variable="x")),
            ]
            # and the log record for two is deleted
            request = respx_mock.calls[-2].request
            assert request.method == "DELETE"
            assert request.url.path == f"/api/api/customentities/~{log_entity}/resource/whatever_two"
            # before the log record for one is deleted last
            request = respx_mock.calls[-1].request
            assert request.method == "DELETE"
            assert request.url.path == f"/api/api/customentities/~{log_entity}/resource/whatever_one"
            # and the url matches the one mocked above


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeResourceRegistry:
    client = create_client(TEST_BASE, "xyz")

    @pytest.fixture()
    def stub_resource_class(self):
        class StubResource(Resource):
            id: str

            def read(self, client, old_state):
                return None

            def create(self, client):
                return None

            def update(self, client, old_state):
                return None

            @staticmethod
            def delete(client, old_state):
                pass

            def deps(self):
                return []

        return StubResource

    @staticmethod
    def test_resource_registry_returns_expected_registrations(stub_resource_class):
        with patch.dict("fbnconfig.resource_abc.RESOURCE_REGISTRY", {}, clear=True) as registry:
            # GIVEN Two resources are registered
            @register_resource(group="SomeGroup", keys=["key1"])
            class KlassResource(stub_resource_class):
                pass

            @register_resource(type_name="DifferentName", group="DifferentGroup", keys=["key2"])
            class MyOtherKlassResource(stub_resource_class):
                pass

            # THEN the resource registry is populated with the expected resources
            assert len(registry) == 2
            assert registry == {
                "KlassResource": {
                    "class": KlassResource,
                    "group": "SomeGroup",
                    "parent": None,
                    "parent_key_mapping": None,
                    "keys": ["key1"],
                },
                "DifferentName": {
                    "class": MyOtherKlassResource,
                    "group": "DifferentGroup",
                    "parent": None,
                    "parent_key_mapping": None,
                    "keys": ["key2"],
                },
            }

    @staticmethod
    def test_get_resource_class_when_unregistered_resource_fails():
        with pytest.raises(RuntimeError) as ex:
            get_resource_class("UnregisteredResource")

        assert "Resource type 'UnregisteredResource' not found in registry" in str(ex.value)

    @staticmethod
    def test_get_resource_type_defaults_to_name():
        resource1 = mock.Mock(spec=Resource)
        resource1.__name__ = "something"
        type = fbnconfig.resource_abc.get_resource_type(resource1)
        assert type == "Mock"


class StubResource(CamelAlias, BaseModel, Resource):
    id: str
    data: str = "default"

    def read(self, client, old_state):
        return None

    def create(self, client):
        return None

    def update(self, client, old_state):
        return None

    @staticmethod
    def delete(client, old_state):
        pass

    def deps(self):
        return []


def ser_stub_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return value


def des_stub_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


StubKey = Annotated[
    StubResource,
    BeforeValidator(des_stub_key),
    PlainSerializer(ser_stub_key),
    WithJsonSchema({
        "type": "object",
        "properties": {"$ref": {"type": "string", "format": "Key.Stub"}},
        "required": ["$ref"]
    }),
]


class FakeResource(BaseModel, CamelAlias, Resource):
    id: str
    with_alias: int | None = None
    depends_on: Sequence[StubKey] | None

    @computed_field(alias="computed")
    def computed(self) -> str:
        return self.id + "computed"

    def read(self, client, old_state):
        return None

    def create(self, client):
        return None

    def update(self, client, old_state):
        return None

    @staticmethod
    def delete(client, old_state):
        pass

    def deps(self):
        return self.depends_on if self.depends_on else []


class TestDumpDeployment:

    def test_dump_one(self):
        # given a resource with an aliased field and a computed one
        # and no dependencies
        fkr = FakeResource(id="one", with_alias=99, depends_on=None)
        deployment = Deployment("dep1", [fkr])
        # when it's dumpted
        serialized = dump_deployment(deployment)
        # then the alias is respected, the computed is omitted
        # and the id is extracted as resourceId
        assert serialized == {
            "deploymentId": "dep1",
            "resources": [{
                "resourceId": "one",
                "resourceType": "FakeResource",
                "dependencies": [],
                "value": {
                    "withAlias": 99,
                }
            }]
        }

    def test_dump_dependents(self):
        stub = StubResource(id="stub-id", data="stubdata")
        fkr = FakeResource(id="fake-id", with_alias=99, depends_on=[stub])
        deployment = Deployment("dep1", [stub, fkr])
        # when it's dumpted
        serialized = dump_deployment(deployment)
        # the dependent resource is first
        assert serialized == {
            "deploymentId": "dep1",
            "resources": [{
                "resourceId": "stub-id",
                "resourceType": "StubResource",
                "dependencies": [],
                "value": {
                    "data": "stubdata",
                }
            }, {
                "resourceId": "fake-id",
                "resourceType": "FakeResource",
                "dependencies": ["stub-id"],
                "value": {
                    "withAlias": 99,
                    "dependsOn": [{"$ref": "stub-id"}]
                }
            }]
        }

    def test_dump_empty_deployment(self):
        # given an empty deployment
        deployment = Deployment("empty-deployment", [])
        # when we dump it
        result = dump_deployment(deployment)
        # then the dump matches expected structure
        expected = {"deploymentId": "empty-deployment", "resources": []}
        assert result == expected

    def test_undump_deployment_with_single_resource(self):
        # given a dumped deployment with one resource
        dump_data = {
            "deploymentId": "test-deployment",
            "resources": [
                {
                    "resourceType": "StubResource",
                    "resourceId": "test-resource",
                    "value": {"data": "test-value"},
                    "dependencies": [],
                }
            ],
        }
        # when we undump it
        with patch.dict(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"StubResource": {"class": StubResource}}
        ):
            result = undump_deployment(dump_data)
        # then we get a deployment with the correct id and resources
        assert result.id == "test-deployment"
        assert len(result.resources) == 1
        resource = result.resources[0]
        assert resource.id == "test-resource"
        assert isinstance(resource, StubResource)
        assert resource.data == "test-value"

    def test_undump_deployment_with_dependent_resources(self):
        # given a dumped deployment with dependent resources
        dump_data = {
            "deploymentId": "test-deployment",
            "resources": [
                {
                    "resourceType": "StubResource",
                    "resourceId": "resource-1",
                    "value": {"data": "value1"},
                    "dependencies": [],
                },
                {
                    "resourceType": "FakeResource",
                    "resourceId": "resource-2",
                    "value": {
                        "withAlias": 87,
                        "dependsOn": [{"$ref": "resource-1"}]
                    },
                    "dependencies": ["resource-1"],
                },
            ],
        }
        # when we undump it
        with patch.dict(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {
                "StubResource": {"class": StubResource},
                "FakeResource": {"class": FakeResource},
            }
        ):
            result = undump_deployment(dump_data)
        # then we get a deployment with both resources
        assert result.id == "test-deployment"
        assert len(result.resources) == 2
        assert result.resources[0].id == "resource-1"
        assert isinstance(result.resources[0], StubResource)
        assert result.resources[0].data == "value1"
        assert isinstance(result.resources[1], FakeResource)
        assert result.resources[1].id == "resource-2"
        assert result.resources[1].with_alias == 87
        assert result.resources[1].depends_on
        assert result.resources[1].depends_on[0] == result.resources[0]

    def test_undump_deployment_with_empty_resources(self):
        # given a dumped deployment with no resources
        dump_data = {"deploymentId": "empty-deployment", "resources": []}
        # when we undump it
        result = undump_deployment(dump_data)
        # then we get an empty deployment
        assert result.id == "empty-deployment"
        assert result.resources == []

    def test_undump_deployment_round_trip(self):
        # given a deployment with resources
        original_resource = StubResource(id="test-resource", data="test-data")
        original_deployment = Deployment("round-trip-test", [original_resource])
        # when we dump and then undump it
        dumped = dump_deployment(original_deployment)
        with patch.dict(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"StubResource": {"class": StubResource}}
        ):
            result = undump_deployment(dumped)
        # then we get back the same deployment
        assert result.id == "round-trip-test"
        assert len(result.resources) == 1
        assert result.resources[0].id == "test-resource"
        assert isinstance(result.resources[0], StubResource)
        assert result.resources[0].data == "test-data"



import os

import pytest
from pytest import fixture

import fbnconfig
from fbnconfig import customentity, datatype, property
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture(scope="module")
def unique_name():
    deployment_name = gen("ce")
    print(f"\nRunning for deployment {deployment_name}...")
    return deployment_name


@fixture(scope="module")
def entity_type(unique_name):
    ce_type = fbnconfig.customentity.EntityTypeResource(
        id="ce2",
        entity_type_name=unique_name,
        display_name="Takeaway menu",
        description="A menu",
        field_schema=[
            fbnconfig.customentity.FieldDefinition(
                name="venueId",
                lifetime=fbnconfig.customentity.LifeTime.PERPETUAL,
                type=fbnconfig.customentity.FieldType.STRING,
                collection_type=fbnconfig.customentity.CollectionType.SINGLE,
                required=True,
            ),
            fbnconfig.customentity.FieldDefinition(
                name="venueOwner",
                lifetime=fbnconfig.customentity.LifeTime.TIMEVARIANT,
                type=fbnconfig.customentity.FieldType.STRING,
                required=True,
            ),
        ],
    )
    return ce_type


@fixture(scope="module")
def identifier_type(unique_name):
    identifier_type = property.DefinitionResource(
        id="exid",
        domain=property.Domain.CustomEntity,
        scope=f"ce-example-{unique_name}",
        code="ceidtype",
        display_name="Example Custom Entity ID",
        data_type_id=datatype.DataTypeRef(
            id="systemstring",
            scope="system",
            code="string",
        ),
        life_time=property.LifeTime.Perpetual,
        constraint_style=property.ConstraintStyle.Identifier,
    )
    return identifier_type


@fixture(scope="module")
def entity_instance(identifier_type, entity_type):
    ce_instance = customentity.EntityResource(
        id="ce1-instance",
        entity_type=entity_type,
        description="An example custom entity instance",
        display_name="Example Instance",
        identifiers=[
            customentity.EntityIdentifier(
                identifier_type=identifier_type,
                identifier_value="ce-example",
            )
        ],
        fields=[
            customentity.EntityField(
                name="venueId",
                value="Example Venue 1",
            ),
            customentity.EntityField(
                name="venueOwner",
                value="Example owner 1",
            )
        ]
    )
    return ce_instance


@fixture(scope="module")
def deployment(unique_name, entity_type, entity_instance):
    dep = fbnconfig.Deployment(unique_name, [entity_type, entity_instance])
    yield dep
    print("\nTearing down deployment but stranding entity type")
    fbnconfig.deployex(fbnconfig.Deployment(unique_name, [entity_type]), lusid_env, token)


def test_create(deployment, identifier_type):
    fbnconfig.deployex(deployment, lusid_env, token)
    res = client.get(f"/api/api/customentities/entitytypes/~{deployment.id}").json()
    assert res is not None
    res = client.get(f"/api/api/customentities/~{deployment.id}/{identifier_type.code}/ce-example",
        params={"identifierScope": identifier_type.scope}).json()
    assert res["fields"][0]["value"] == "Example Venue 1"


def test_nochange(deployment):
    fbnconfig.deployex(deployment, lusid_env, token)
    update = fbnconfig.deploy(deployment, lusid_env, token)
    assert [a.change for a in update] == ["nochange", "attach", "nochange", "nochange"]


def test_update_entity(deployment, entity_type, identifier_type):
    # given an existing deployment
    fbnconfig.deployex(deployment, lusid_env, token)
    # and an instance with some mnodified fields
    ce_instance = customentity.EntityResource(
        id="ce1-instance",
        entity_type=entity_type,
        description="An example custom entity instance",
        display_name="Example Instance",
        identifiers=[
            customentity.EntityIdentifier(
                identifier_type=identifier_type,
                identifier_value="ce-example",
            )
        ],
        fields=[
            customentity.EntityField(
                name="venueId",
                value="Different Venue 1",
            ),
            customentity.EntityField(
                name="venueOwner",
                value="Example owner 1",
            )
        ]
    )
    # when we deploy over the top
    mod_deploy = fbnconfig.Deployment(deployment.id, [entity_type, ce_instance])
    fbnconfig.deployex(mod_deploy, lusid_env, token)
    # then the field has changed
    res = client.get(f"/api/api/customentities/~{deployment.id}/{identifier_type.code}/ce-example",
        params={"identifierScope": identifier_type.scope}).json()
    assert res["fields"][0]["value"] == "Different Venue 1"


def test_update_entity_identifier(deployment, entity_type, identifier_type):
    # given an existing deployment
    fbnconfig.deployex(deployment, lusid_env, token)
    # and an instance with some mnodified fields
    ce_instance = customentity.EntityResource(
        id="ce1-instance",
        entity_type=entity_type,
        description="An example custom entity instance",
        display_name="Example Instance",
        identifiers=[
            customentity.EntityIdentifier(
                identifier_type=identifier_type,
                identifier_value="different-example",
            )
        ],
        fields=[
            customentity.EntityField(
                name="venueId",
                value="Different Venue 1",
            ),
            customentity.EntityField(
                name="venueOwner",
                value="Example owner 1",
            )
        ]
    )
    # when we deploy over the top
    mod_deploy = fbnconfig.Deployment(deployment.id, [entity_type, ce_instance])
    fbnconfig.deployex(mod_deploy, lusid_env, token)
    # then the field has changed
    res = client.get("/api/api/customentities/"
        f"~{deployment.id}/{identifier_type.code}/different-example",
        params={"identifierScope": identifier_type.scope}).json()
    assert res["fields"][0]["value"] == "Different Venue 1"


@pytest.mark.asyncio
async def test_entity_list(deployment, unique_name):
    # GIVEN Entity is deployed
    fbnconfig.deployex(deployment, lusid_env, token)
    deployment_name = unique_name
    entity_type = f"~{deployment_name}"
    identifier_scope = f"ce-example-{deployment_name}"
    identifier_code = "ceidtype"
    identifier_value = "ce-example"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await customentity.EntityResource.list(
        client=client,
        query_params={},
        path_params={"entityType": entity_type}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "entityType": entity_type,
                    "identifierType": identifier_code,
                    "identifierValue": identifier_value,
                    "identifierScope": identifier_scope,
                },
                "value": {
                    "id": (f"entity_{entity_type}_{identifier_code}_"
                           f"{identifier_value}_{identifier_scope}".lower()),
                    "entityType": {
                        "$ref": f"entity-type_{entity_type}".lower()
                    },
                    "description": "An example custom entity instance",
                    "displayName": "Example Instance",
                    "identifiers": [
                        {
                            "identifierType": {
                                "$ref": f"identifier-type_{identifier_code}".lower()
                            },
                            "identifierValue": identifier_value,
                            "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                            "effectiveTo": "9999-12-31T23:59:59.9999999+00:00"
                        }
                    ],
                    "fields": [
                        {
                            "name": "venueId",
                            "value": "Example Venue 1",
                            "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                            "effectiveTo": "9999-12-31T23:59:59.9999999+00:00"
                        },
                        {
                            "name": "venueOwner",
                            "value": "Example owner 1",
                            "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                            "effectiveTo": "9999-12-31T23:59:59.9999999+00:00"
                        }
                    ],
                    "properties": {},
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_entity_fetch(deployment, unique_name):
    # GIVEN Entity is deployed
    fbnconfig.deployex(deployment, lusid_env, token)
    deployment_name = unique_name
    entity_type = f"~{deployment_name}"
    identifier_scope = f"ce-example-{deployment_name}"
    identifier_code = "ceidtype"
    identifier_value = "ce-example"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await customentity.EntityResource.fetch(
        client=client,
        keys={
            "entityType": entity_type,
            "identifierType": identifier_code,
            "identifierValue": identifier_value,
            "identifierScope": identifier_scope,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"entity_{entity_type}_{identifier_code}_{identifier_value}_{identifier_scope}".lower(),
        "entityType": {
            "$ref": f"entity-type_{entity_type}".lower()
        },
        "description": "An example custom entity instance",
        "displayName": "Example Instance",
        "identifiers": [
            {
                "identifierType": {
                    "$ref": f"identifier-type_{identifier_code}".lower()
                },
                "identifierValue": identifier_value,
                "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                "effectiveTo": "9999-12-31T23:59:59.9999999+00:00"
            }
        ],
        "fields": [
            {
                "name": "venueId",
                "value": "Example Venue 1",
                "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                "effectiveTo": "9999-12-31T23:59:59.9999999+00:00"
            },
            {
                "name": "venueOwner",
                "value": "Example owner 1",
                "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                "effectiveTo": "9999-12-31T23:59:59.9999999+00:00"
            }
        ],
        "properties": {},
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_entity_type_list_with_filter(deployment, unique_name):
    # GIVEN CustomEntity is deployed
    fbnconfig.deployex(deployment, lusid_env, token)
    entity_type = f"~{unique_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await customentity.EntityTypeResource.list(
        client=client,
        query_params={"filter": f"entityTypeName eq '{unique_name}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "entityType": entity_type,
                },
                "value": {
                    "id": f"entity-type_{entity_type}".lower(),
                    "entityTypeName": unique_name,
                    "displayName": "Takeaway menu",
                    "description": "A menu",
                    "fieldSchema": [
                        {
                            "lifetime": "Perpetual",
                            "name": "venueId",
                            "type": "String",
                            "required": True
                        },
                        {
                            "lifetime": "TimeVariant",
                            "name": "venueOwner",
                            "type": "String",
                            "required": True
                        }
                    ]
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_entity_type_fetch(deployment, unique_name):
    # GIVEN CustomEntity is deployed
    fbnconfig.deployex(deployment, lusid_env, token)
    entity_type = f"~{unique_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await customentity.EntityTypeResource.fetch(
        client=client,
        keys={
            "entityType": entity_type
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"entity-type_{entity_type}".lower(),
        "entityTypeName": unique_name,
        "displayName": "Takeaway menu",
        "description": "A menu",
        "fieldSchema": [
            {
                "lifetime": "Perpetual",
                "name": "venueId",
                "type": "String",
                "required": True
            },
            {
                "lifetime": "TimeVariant",
                "name": "venueOwner",
                "type": "String",
                "required": True
            }
        ]
    }
    assert response == expected_response



import http
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator
from pydantic_core import InitErrorDetails
from pydantic_core.core_schema import ValidationInfo
from starlette.requests import Request
from starlette.responses import JSONResponse

import fbnconfig
from fbnconfig.exceptions import NotFoundError
from fbnconfig.resource_abc import RESOURCE_REGISTRY, Ref, Resource


async def get_metadata(_: Request):
    def is_resource(entry: dict) -> bool:
        cls = entry["class"]
        return issubclass(cls, Resource)

    def is_ref(entry: dict) -> bool:
        cls = entry["class"]
        return issubclass(cls, Ref)

    def has_method(cls: type, method_name: str) -> bool:
        return callable(getattr(cls, method_name, None))

    def has_ref(cls: type) -> bool:
        ref_name = cls.__name__.removesuffix("Resource") + "Ref"
        return ref_name in RESOURCE_REGISTRY and is_ref(RESOURCE_REGISTRY[ref_name])

    def get_resource_metadata(entry: dict) -> dict:
        cls = entry["class"]
        group = entry["group"]
        parent = entry["parent"]
        keys = entry["keys"]
        return {
            "resource_type": cls.resource_type,
            "group": group,
            "parent": parent.__name__ if parent else None,
            "list_keys": parent.keys if parent else [],
            "fetch_keys": keys,
            "filterable_columns": cls.filterable_fields,
            "supports_list": has_method(cls, "list"),
            "supports_fetch": has_method(cls, "fetch"),
            "supports_ref": has_ref(cls)
        }

    return JSONResponse([
        get_resource_metadata(entry)
        for entry in RESOURCE_REGISTRY.values()
        if is_resource(entry)
    ])


async def get_list(request: Request):
    class ListRequestParams(BaseModel):
        resource: str = Field(pattern=r"^[A-Za-z]+$")
        limit: int | None = Field(50, ge=1)
        page: str | None = None
        filter: str | None = None
        keys: dict[str, Any] | None = None
        model_config = ConfigDict(extra="ignore")

    try:
        body = await request.json()
        request_params = ListRequestParams(**(body | request.path_params))
    except ValidationError as e:
        return JSONResponse(
            status_code=http.HTTPStatus.UNPROCESSABLE_ENTITY,
            content=e.errors(),
        )

    resource = get_resource(request_params.resource)
    resource_class = resource["class"]
    if not has_method(resource_class, "list"):
        return JSONResponse(
            status_code=http.HTTPStatus.NOT_IMPLEMENTED,
            content={"detail": "Method not implemented for this resource"}
        )

    client = fbnconfig.create_client(request.state.lusid_env, request.state.token, is_async=True)
    data = await resource_class.list(
        client=client,
        query_params=request_params.model_dump(exclude_none=True, exclude={"resource", "keys"}),
        path_params=request_params.keys
    )

    return JSONResponse(data)


async def get_fetch(request: Request):
    class FetchRequestParams(BaseModel):
        resource: str = Field(pattern=r"^[A-Za-z]+$")

        model_config = ConfigDict(extra="ignore")

    class ResourceKeyParams(BaseModel):
        keys: dict[str, Any]

        @field_validator("keys")
        @classmethod
        def check_mandatory_keys(cls, keys: dict, info: ValidationInfo):
            if info.context is None or "mandatory_keys" not in info.context:
                raise ValueError("Validation context is missing")

            mandatory_keys = set(info.context["mandatory_keys"])
            provided_keys = set(keys.keys())

            missing_keys = mandatory_keys - provided_keys
            extra_keys = provided_keys - mandatory_keys

            errors = []

            for key in missing_keys:
                errors.append(
                    InitErrorDetails(type="missing", loc=(key,), input=keys,)
                )

            for key in extra_keys:
                errors.append(
                    InitErrorDetails(type="extra_forbidden", loc=(key,), input=key,)
                )

            if errors:
                raise ValidationError.from_exception_data(cls.__name__, errors)

            return keys

    try:
        request_params = FetchRequestParams(**request.path_params)
    except ValidationError as e:
        return JSONResponse(
            status_code=http.HTTPStatus.UNPROCESSABLE_ENTITY,
            content=e.errors(),
        )

    resource = get_resource(request_params.resource)
    resource_class = resource["class"]
    if not has_method(resource_class, "fetch"):
        return JSONResponse(
            status_code=http.HTTPStatus.NOT_IMPLEMENTED,
            content={"detail": "Method not implemented for this resource"}
        )

    try:
        body = await request.json()
        params = ResourceKeyParams.model_validate(
            body, context={"mandatory_keys": resource["keys"]}
        )
    except ValidationError as e:
        return JSONResponse(
            status_code=http.HTTPStatus.UNPROCESSABLE_ENTITY,
            content=e.errors(),
        )

    client = fbnconfig.create_client(request.state.lusid_env, request.state.token, is_async=True)
    data = await resource_class.fetch(client=client, keys=params.keys)

    return JSONResponse(data)


def get_resource(resource_type: str) -> dict[str, Any]:
    """
    Get the resource entry from the registry by its resource_type.
    """
    resource_object = RESOURCE_REGISTRY.get(resource_type)
    if resource_object is None:
        raise NotFoundError("resource")
    return resource_object


def has_method(cls: type, method_name: str) -> bool:
    return callable(getattr(cls, method_name, None))



import datetime as dt

from pydantic import BaseModel

from fbnconfig.coretypes import IsoDateTime


class DateModel(BaseModel):
    """Minimal model for testing IsoDateTime serialization/deserialization."""
    value: IsoDateTime


class DescribeIsoDateTime:
    """Tests for IsoDateTime (des_isodate + ser_isodate) through model validation and serialization."""

    def test_validate_plain_date(self):
        model = DateModel.model_validate({"value": dt.date(2024, 6, 15)})
        assert model.value == dt.date(2024, 6, 15)
        assert type(model.value) is dt.date

    def test_validate_datetime_normalized_to_utc(self):
        value = dt.datetime(2024, 6, 15, 10, 30, 0, tzinfo=dt.timezone(dt.timedelta(hours=5)))
        model = DateModel.model_validate({"value": value})
        assert model.value == dt.datetime(2024, 6, 15, 5, 30, 0, tzinfo=dt.timezone.utc)

    def test_validate_iso_string_with_timezone(self):
        model = DateModel.model_validate({"value": "2024-06-15T10:30:00+00:00"})
        assert model.value == dt.datetime(2024, 6, 15, 10, 30, 0, tzinfo=dt.timezone.utc)

    def test_validate_iso_string_without_timezone(self):
        model = DateModel.model_validate({"value": "2024-06-15T10:30:00"})
        assert model.value == dt.datetime(2024, 6, 15, 10, 30, 0)

    def test_validate_date_only_string(self):
        model = DateModel.model_validate({"value": "2024-06-15"})
        assert isinstance(model.value, dt.date)
        assert model.value.year == 2024
        assert model.value.month == 6
        assert model.value.day == 15

    def test_serialize_utc_datetime(self):
        model = DateModel.model_validate({"value": "2024-06-15T10:30:00+00:00"})
        result = model.model_dump(mode="json")
        assert result == {"value": "2024-06-15T10:30:00+00:00"}

    def test_serialize_date(self):
        model = DateModel.model_validate({"value": dt.date(2024, 6, 15)})
        result = model.model_dump(mode="json")
        assert result == {"value": "2024-06-15"}



from fbnconfig import Deployment
from fbnconfig.configuration import SystemConfigResource

"""
An example configuration for defining system configuration entities.
The script configures the following entities:
- System configuration

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01768/
https://support.lusid.com/knowledgebase/article/KA-01799/
"""


def configure(env):
    validate_instruments = SystemConfigResource(
        id="validate-instruments",
        code="TransactionBooking",
        key="ValidateInstruments",
        value="True",
        description="Sets validation that instruments exist when booking transactions",
        default_value=False,
    )

    validate_txn_types = SystemConfigResource(
        id="validate-txn-types",
        code="TransactionBooking",
        key="ValidateTransactionTypes",
        value="False",
        description="Sets validation that transaction types exist when booking transactions",
        default_value=True,
    )

    return Deployment("system-configuration", [validate_instruments, validate_txn_types])




from typing import Annotated, Mapping, Self, Sequence

from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, model_validator

from .coretypes import IsoDateTime
from .property import PropertyKey
from .resource_abc import CamelAlias


class MetricValue(CamelAlias, BaseModel):
    value: float
    unit: str | None = None


# The inner part of a property value for when it is used inside a structure
# that has the property definition
class PropertyValueInner(CamelAlias, BaseModel):
    label_value: str | None = None  # noqa: N815
    metric_value: MetricValue | None = None  # noqa: N815
    label_set_value: Sequence[str] | None = None  # noqa: N815

    @model_validator(mode="after")
    def validate_one_value_exists(self):
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")
        return self


class PropertyValue(CamelAlias, BaseModel):
    property_key: PropertyKey = Field(serialization_alias="propertyKey")
    label_value: str | None = Field(default=None, serialization_alias="labelValue")
    metric_value: MetricValue | None = Field(default=None, serialization_alias="metricValue")
    label_set_value: Sequence[str] | None = Field(default=None, serialization_alias="labelSetValue")
    effective_from: IsoDateTime | None = None
    effective_until: IsoDateTime | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data
        style = info.context.get("style", "api") if info.context else "api"
        if style == "api" and data.get("value", None):
            return data | data["value"]
        return data

    @model_validator(mode="after")
    def validate_one_value_exists(self) -> Self:
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")

        return self


def ser_properties_as_dict(values: Sequence[PropertyValue], info):
    style = info.context.get("style", "api") if info.context else "api"
    if style == "dump":
        return values
    property_dict = {}
    for prop in values:
        value = {}
        if prop.label_value:
            value["labelValue"] = prop.label_value
        elif prop.metric_value:
            value["metricValue"] = prop.metric_value
        elif prop.label_set_value:
            value["labelSetValue"] = prop.label_set_value
        key = f"{prop.property_key.domain.value}/{prop.property_key.scope}/{prop.property_key.code}"
        property_dict[key] = {"key": key, "value": value}

        if prop.effective_from is not None:
            property_dict[key]["effectiveFrom"] = prop.effective_from

        if prop.effective_until is not None:
            property_dict[key]["effectiveUntil"] = prop.effective_until

    return property_dict


def des_properties_dict(data, info) -> Sequence[PropertyValue] | Mapping | None:
    if not isinstance(data, dict) or data is None:
        return data
    style = info.context.get("style", "api") if info.context else "api"
    if style == "api" and isinstance(data, dict):
        data = [value | {"propertyKey": value["key"]} for value in data.values()]
    return data


PropertyValueDict = Annotated[
    Sequence[PropertyValue],
    BeforeValidator(des_properties_dict), PlainSerializer(ser_properties_as_dict)
]



import hashlib
import json
import pathlib
from types import SimpleNamespace
from unittest import mock

import httpx
import pytest
from pydantic import ValidationError

from fbnconfig import drive


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeFolderRef:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url)

    def test_attach(self, respx_mock):
        respx_mock.post("/drive/api/search/").mock(
            return_value=httpx.Response(200, json={"values": [{"id": 23456}]})
        )
        ref = drive.FolderRef(id="a", folder_path="/robTest99")
        ref.attach(self.client)
        assert ref.drive_id == 23456

    def test_attach_root(self):
        # given a ref to the root path
        ref = drive.FolderRef(id="abc", folder_path="/")
        # when we attach it
        ref.attach(self.client)
        # then no http requests are made (you can't search for root with the api)
        # and the driveId is root
        assert ref.drive_id == "/"


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeFolder:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url)

    def test_ctor_invalid_parent(self):
        try:
            drive.FolderResource(id="id", name="name", parent=None)  # type: ignore
        except ValidationError:
            return
        assert False

    def test_ctor_with_root(self):
        f = drive.FolderResource(id="id", name="name", parent=drive.root)
        assert f

    def test_create_in_root(self, respx_mock):
        respx_mock.post("/drive/api/folders").mock(return_value=httpx.Response(200, json={"id": 23456}))
        client = self.client
        # given nothing really
        # when we create a folder at the root level
        sut = drive.FolderResource(id="anid", name="new_folder_name", parent=drive.root)
        state = sut.create(client)
        # then the state is returned
        assert state["name"] == "new_folder_name"
        assert state["driveId"] == 23456
        assert state["parentId"] == "/"
        assert str(sut.path()) == "/new_folder_name"
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/drive/api/folders"
        assert json.loads(request.content) == {"path": "/", "name": "new_folder_name"}

    def test_create_in_folder(self, respx_mock):
        client = self.client
        parent = drive.FolderResource(id="parentid", name="parentname", parent=drive.root)
        respx_mock.post("/drive/api/folders").mock(
            side_effect=[httpx.Response(200, json={"id": 1234}), httpx.Response(200, json={"id": 23456})]
        )
        parent.create(client)
        sut = drive.FolderResource(id="anid", name="new_folder_name", parent=parent)
        state = sut.create(client)
        assert state["name"] == "new_folder_name", "should have child name"
        assert state["driveId"] == 23456, "should get child driveId"
        assert state["parentId"] == 1234, "should capture parent driveId"
        assert str(sut.path()) == "/parentname/new_folder_name"
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/drive/api/folders"
        assert json.loads(request.content) == {"path": "/parentname", "name": "new_folder_name"}

    @pytest.fixture
    def existing_structure(self, respx_mock):
        # A structure which starts as
        # root/
        #    parentname/               [driveid: 1234]
        #       subfolder2             [driveid: 23456]
        #    other/                    [driveid: 7654]
        client = self.client
        respx_mock.post("/drive/api/folders").mock(
            side_effect=[
                httpx.Response(200, json={"id": "1234"}),
                httpx.Response(200, json={"id": "23456"}),
                httpx.Response(200, json={"id": "7654"}),
            ]
        )
        parent = drive.FolderResource(id="parentid", name="parentname", parent=drive.root)
        parent.create(client)
        folder = drive.FolderResource(id="anid", name="subfolder2", parent=parent)
        folder.create(client)
        other = drive.FolderResource(id="other", name="other", parent=drive.root)
        return SimpleNamespace(parent=parent, subfolder=folder, other=other)

    def test_delete(self, respx_mock, existing_structure):
        respx_mock.delete("/drive/api/folders/23456").mock(
            side_effect=[httpx.Response(200, json={"id": "23456"})]
        )

        # given an existing folder
        old_state = SimpleNamespace(id="anid", name="subfolder2", driveId="23456", parentId="1234")
        # when we delete it
        sut = drive.FolderResource(id="anid", name="subfolder2", parent=existing_structure.parent)
        sut.delete(self.client, old_state)
        # then a delete request is sent
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url == f"{self.base_url}/drive/api/folders/23456"

    def test_update_nochange(self, existing_structure):
        # given an existing state
        old_state = SimpleNamespace(id="anid", name="subfolder2", driveId="23456", parentId="1234")
        # when we update the folder with the same values as the state
        sut = drive.FolderResource(id="anid", name="subfolder2", parent=existing_structure.parent)
        new_state = sut.update(self.client, old_state)
        # then new  state is None indicating that no change was made
        assert new_state is None
        # and no http request was made (no mock!)

    def test_update_name(self, respx_mock, existing_structure):
        respx_mock.put("/drive/api/folders/23456").mock(
            side_effect=[httpx.Response(200, json={"id": "23456"})]
        )
        # given the folder exists with name 'another'
        old_state = SimpleNamespace(id="anid", name="subfolder2", driveId="23456", parentId="1234")
        # when it is updated with name subfolder2
        sut = drive.FolderResource(id="anid", name="anotherName", parent=existing_structure.parent)
        new_state = sut.update(self.client, old_state)
        # then a put request is made with the new name and the existing driveId
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/drive/api/folders/23456"
        assert json.loads(request.content) == {"path": "/parentname", "name": "anotherName"}
        # and the new state is returned
        assert new_state == {"id": "anid", "name": "anotherName", "driveId": "23456", "parentId": "1234"}

    def test_update_parent(self, respx_mock, existing_structure):
        respx_mock.put("/drive/api/folders/23456").mock(
            side_effect=[httpx.Response(200, json={"id": "23456"})]
        )
        # given a state  /other/subfolder2     [other is id:7654]
        old_state = SimpleNamespace(id="anid", name="subfolder2", driveId="23456", parentId="7654")
        # when we update it's parent to be /parentname  [1234]
        sut = drive.FolderResource(id="anid", name="subfolder2", parent=existing_structure.parent)
        new_state = sut.update(self.client, old_state)
        # then a put request is made to move the subject
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/drive/api/folders/23456"
        assert json.loads(request.content) == {"path": "/parentname", "name": "subfolder2"}
        # and the new state is returned
        assert new_state == {"id": "anid", "name": "subfolder2", "driveId": "23456", "parentId": "1234"}

    def test_dump_simple_folder(self):
        # given a simple folder resource
        sut = drive.FolderResource(
            id="dump-folder",
            name="DumpFolder",
            parent=drive.root
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then all fields are included (no excludes)
        expected = {
            "name": "DumpFolder",
            "parent": {"$ref": "drive_root"}
        }
        assert result == expected

    def test_dump_nested_folder(self):
        # given a nested folder structure
        parent_folder = drive.FolderResource(
            id="parent-folder",
            name="ParentFolder",
            parent=drive.root
        )
        sut = drive.FolderResource(
            id="child-folder",
            name="ChildFolder",
            parent=parent_folder
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then parent is referenced by $ref
        expected = {
            "name": "ChildFolder",
            "parent": {"$ref": "parent-folder"}
        }
        assert result == expected

    def test_undump_simple_folder(self):
        # given dump data
        data = {
            "name": "UndumpFolder",
            "parent": {"$ref": "drive_root"}
        }
        # mock the $refs context with root reference
        refs = {"drive_root": drive.root}
        # when we undump it with id from context
        result = drive.FolderResource.model_validate(
            data, context={"style": "dump", "id": "folder_id", "$refs": refs}
        )
        # then it's correctly populated including id from context
        assert result.id == "folder_id"
        assert result.name == "UndumpFolder"
        assert result.parent == drive.root

    def test_undump_nested_folder(self):
        # given dump data with parent folder reference
        data = {
            "name": "UndumpChildFolder",
            "parent": {"$ref": "parent-folder-id"}
        }
        # mock parent folder
        parent_folder = drive.FolderResource(
            id="parent-folder-id",
            name="ParentFolder",
            parent=drive.root
        )
        refs = {"parent-folder-id": parent_folder}
        # when we undump it
        result = drive.FolderResource.model_validate(
            data, context={"style": "dump", "id": "child_folder_id", "$refs": refs}
        )
        # then it's correctly populated
        assert result.id == "child_folder_id"
        assert result.name == "UndumpChildFolder"
        assert result.parent == parent_folder

    def test_from_api(self, testdata_loader):
        input_data = testdata_loader(
            "folder_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "folder_resource",
            "list_expected_result"
        )["values"][0]["value"]
        result_data = drive.FolderResource.from_api(input_data)
        assert result_data == expected_result

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        input_data = testdata_loader(
            "folder_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "folder_resource",
            "list_expected_result"
        )
        respx_mock.get(drive.FolderResource.list_endpoint()).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await drive.FolderResource.list(client, query_params={})
            assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        input_data = testdata_loader(
            "folder_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "folder_resource",
            "list_expected_result"
        )["values"][0]["value"]
        keys = {"driveId": "cut-label_lum-cut-tcode-6"}
        respx_mock.get(drive.FolderResource.fetch_endpoint().format(**keys)).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await drive.FolderResource.fetch(client, keys=keys)
        assert result == expected_result


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeFile:
    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url)

    def hashit(self, content):
        encoded = content.encode() if isinstance(content, str) else content
        return hashlib.sha256(encoded).hexdigest()

    @pytest.fixture
    def existing_structure(self, respx_mock):
        # A structure which starts as
        # root/
        #    parentname/               [driveid: 1234]
        #       subfolder2/            [driveid: 23456]
        #          afile.txt           [driveid: 6543]
        #    other/                    [driveid: 7654]
        client = self.client
        parent = drive.FolderResource(id="parentid", name="parentname", parent=drive.root)
        respx_mock.post("/drive/api/folders").mock(
            side_effect=[
                httpx.Response(200, json={"id": "1234"}),
                httpx.Response(200, json={"id": "23456"}),
                httpx.Response(200, json={"id": "7654"}),
            ]
        )
        parent.create(client)
        folder = drive.FolderResource(id="anid", name="subfolder2", parent=parent)
        folder.create(client)
        other = drive.FolderResource(id="other", name="other", parent=drive.root)
        other.create(client)
        file = drive.FileResource(id="afile", name="afile.txt", folder=folder, content="some content")
        return SimpleNamespace(parent=parent, subfolder=folder, other=other, file=file)

    def test_ctor_throws_with_non_folder(self):
        not_a_folder = {}
        with pytest.raises(ValidationError):
            drive.FileResource(id="id", name="name", folder=not_a_folder, content="content")  # type: ignore

    def test_ctor_works_with_folder(self):
        folder = drive.FolderResource(id="folder_id", name="folder", parent=drive.root)
        f = drive.FileResource(id="id", name="name", folder=folder, content="hello")
        assert f

    def test_delete(self, respx_mock, existing_structure):
        respx_mock.delete("/drive/api/files/6543").mock(side_effect=[httpx.Response(200, json={})])
        # given the existing file: afile.txt
        # when we delete it
        old_state = SimpleNamespace(id="afile", name="afile.txt", driveId="6543", parentId="123456")
        sut = drive.FileResource(
            id="afile", name="afile.txt", folder=existing_structure.subfolder, content="fake"
        )
        sut.delete(self.client, old_state)
        # then a delete request is made
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url == f"{self.base_url}/drive/api/files/6543"

    def test_create_str_content(self, respx_mock, existing_structure):
        respx_mock.post("/drive/api/files").mock(side_effect=[httpx.Response(200, json={"id": "8888"})])
        # given an existing folder, subfolder2 [23456]
        parent = existing_structure.subfolder
        # when we create a file with string content
        sut = drive.FileResource(id="bfile", name="poem.txt", folder=parent, content="banjo poem")
        new_state = sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/drive/api/files"
        assert request.headers["x-lusid-drive-filename"] == "poem.txt"
        assert request.headers["x-lusid-drive-path"] == "/parentname/subfolder2"
        assert request.headers["content-type"] == "application/octet-stream"
        # and the new state is created
        assert new_state == {
            "id": "bfile",
            "name": "poem.txt",
            "driveId": "8888",
            "parentId": "23456",
            "content_hash": self.hashit("banjo poem"),
        }

    def test_move_folders(self, respx_mock, existing_structure):
        respx_mock.put("/drive/api/files/6543").mock(side_effect=[httpx.Response(200, json={})])
        # given an existing file parentname/subfolder2/afile.txt
        old_parent = existing_structure.subfolder
        # when we change the parent to /other [id:7654]
        old_state = SimpleNamespace(
            id="cfile",
            name="afile.txt",
            driveId="6543",
            parentId=old_parent.id,
            content_hash=self.hashit("jam poem"),
        )
        sut = drive.FileResource(
            id="cfile", name="afile.txt", folder=existing_structure.other, content="jam poem"
        )
        new_state = sut.update(self.client, old_state)
        # then an out request is made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/drive/api/files/6543"
        assert json.loads(request.content) == {"path": "/other", "name": "afile.txt"}
        # and the new state is created
        assert new_state == {
            "id": "cfile",
            "name": "afile.txt",
            "driveId": "6543",
            "parentId": "7654",
            "content_hash": self.hashit("jam poem"),
        }

    def test_rename_file(self, respx_mock, existing_structure):
        respx_mock.put("/drive/api/files/6543").mock(side_effect=[httpx.Response(200, json={})])
        # given an existing file parentname/subfolder2/afile.txt
        parent = existing_structure.subfolder
        # when we rename the file but retain the parent
        old_state = SimpleNamespace(
            id="cfile",
            name="afile.txt",
            driveId="6543",
            parentId=parent.id,
            content_hash=self.hashit("bear tail"),
        )
        sut = drive.FileResource(id="cfile", name="renamed.sql", folder=parent, content="bear tail")
        new_state = sut.update(self.client, old_state)
        # then a out request is made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/drive/api/files/6543"
        assert json.loads(request.content) == {"path": "/parentname/subfolder2", "name": "renamed.sql"}
        # and the new state is created
        assert new_state == {
            "id": "cfile",
            "name": "renamed.sql",
            "driveId": "6543",
            "parentId": parent.drive_id,
            "content_hash": self.hashit("bear tail"),
        }

    def test_change_content(self, respx_mock, existing_structure):
        respx_mock.put("/drive/api/files/6543/contents").mock(side_effect=[httpx.Response(200, json={})])
        # given an existing file parentname/subfolder2/afile.txt
        parent = existing_structure.subfolder
        # when we update the content
        old_state = SimpleNamespace(
            id="cfile",
            name="afile.txt",
            driveId="6543",
            parentId=parent.drive_id,
            content_hash=self.hashit("cow bell"),
        )
        sut = drive.FileResource(id="cfile", name="afile.txt", folder=parent, content="pig ears")
        new_state = sut.update(self.client, old_state)
        # then a put request is made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url == f"{self.base_url}/drive/api/files/6543/contents"
        assert request.headers["content-type"] == "application/octet-stream"
        assert request.content == b"pig ears"
        # and the new state is created
        assert new_state == {
            "id": "cfile",
            "name": "afile.txt",
            "driveId": "6543",
            "parentId": parent.drive_id,
            "content_hash": self.hashit("pig ears"),
        }

    def test_ctor_with_file_content(self, existing_structure):
        mock_open = mock.mock_open(read_data=b"juicy\njuicy")
        with mock.patch("builtins.open", mock_open) as m:
            sut = drive.FileResource(
                id="from-file",
                name="tomato.txt",
                folder=existing_structure.parent,
                content_path=pathlib.PurePath("/foo.txt"),
            )
            assert sut.get_content() == b"juicy\njuicy"
        m.assert_called_with(pathlib.PurePosixPath("/foo.txt"), "rb")

    def test_create_with_file_content(self, respx_mock, existing_structure):
        respx_mock.post("/drive/api/files").mock(side_effect=[httpx.Response(200, json={"id": "999"})])
        mock_open = mock.mock_open(read_data=b"juicy\njuicy")
        with mock.patch("builtins.open", mock_open):
            sut = drive.FileResource(
                id="from-file",
                name="tomato.txt",
                folder=existing_structure.subfolder,
                content_path=pathlib.PurePath("/foo.txt"),
            )
            new_state = sut.create(self.client)
        # then a post request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url == f"{self.base_url}/drive/api/files"
        assert request.headers["x-lusid-drive-filename"] == "tomato.txt"
        assert request.headers["x-lusid-drive-path"] == "/parentname/subfolder2"
        assert request.headers["content-type"] == "application/octet-stream"
        assert request.content == b"juicy\njuicy"
        # and the new state is created
        assert new_state == {
            "id": "from-file",
            "name": "tomato.txt",
            "driveId": "999",
            "parentId": "23456",
            "content_hash": self.hashit(b"juicy\njuicy"),
        }

    def test_dump_simple_file(self):
        # given a simple file resource
        folder = drive.FolderResource(
            id="test-folder",
            name="TestFolder",
            parent=drive.root
        )
        sut = drive.FileResource(
            id="dump-file",
            name="DumpFile.txt",
            folder=folder,
            content="test content"
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then all fields are included (no excludes)
        expected = {
            "name": "DumpFile.txt",
            "content": "test content",
            "folder": {"$ref": "test-folder"}
        }
        assert result == expected

    def test_dump_file_bytes_content(self):
        # given a file resource with bytes content
        folder = drive.FolderResource(
            id="test-folder",
            name="TestFolder",
            parent=drive.root
        )
        sut = drive.FileResource(
            id="dump-file-bytes",
            name="DumpFile.bin",
            folder=folder,
            content=b"binary content"
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then bytes content is converted to string in json mode
        expected = {
            "name": "DumpFile.bin",
            "content": "binary content",
            "folder": {"$ref": "test-folder"}
        }
        assert result == expected

    def test_undump_simple_file(self):
        # given dump data
        data = {
            "name": "UndumpFile.txt",
            "content": "undump content",
            "folder": {"$ref": "test-folder-id"}
        }
        # mock folder reference
        folder = drive.FolderResource(
            id="test-folder-id",
            name="TestFolder",
            parent=drive.root
        )
        refs = {"test-folder-id": folder}
        # when we undump it with id from context
        result = drive.FileResource.model_validate(
            data, context={"style": "dump", "id": "file_id", "$refs": refs}
        )
        # then it's correctly populated including id from context
        assert result.id == "file_id"
        assert result.name == "UndumpFile.txt"
        assert result.content == "undump content"
        assert result.folder == folder

    def test_undump_file_bytes_content(self):
        # given dump data with bytes content
        data = {
            "name": "UndumpFile.bin",
            "content": b"undump binary content",
            "folder": {"$ref": "test-folder-id"}
        }
        # mock folder reference
        folder = drive.FolderResource(
            id="test-folder-id",
            name="TestFolder",
            parent=drive.root
        )
        refs = {"test-folder-id": folder}
        # when we undump it
        result = drive.FileResource.model_validate(
            data, context={"style": "dump", "id": "file_bytes_id", "$refs": refs}
        )
        # then it's correctly populated
        assert result.id == "file_bytes_id"
        assert result.name == "UndumpFile.bin"
        assert result.content == b"undump binary content"
        assert result.folder == folder

    def test_from_api(self, testdata_loader):
        # GIVEN a File resource
        input_data = testdata_loader(
            "file_resource",
            "list_mock_data"
        )[0]
        expected_result = testdata_loader(
            "file_resource",
            "list_expected_result"
        )["values"][0]
        # AND the required keys are populated
        input_data["driveId"] = "driveId"
        # WHEN the from_api method is called for that resource
        result = drive.FileResource.from_api(input_data)
        # THEN the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        # GIVEN a File resource
        input_data = testdata_loader(
            "file_resource",
            "list_mock_data"
        )[0]
        expected_result = testdata_loader(
            "file_resource",
            "list_expected_result"
        )["values"][0]
        # AND the resource exists in LUSID
        keys = {
            "driveId": "driveId",
        }
        url = drive.FileResource.fetch_endpoint().format(**keys)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the fetch method is called for that resource
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await drive.FileResource.fetch(client, keys=keys)
        # THEN the response is as expected
        assert result == expected_result
        # AND the correct request was made to LUSID
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == url
        assert request.url.query == b""



import os
from types import SimpleNamespace

import pytest

import fbnconfig
from fbnconfig import Deployment, identity
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


def resources(deployment_name):
    return [
        identity.ApplicationResource(
            id="test_app",
            client_id=f"inttest-app-{deployment_name}",
            display_name=f"Application{deployment_name}",
            type=identity.ApplicationType.NATIVE,
        )
    ]


# this cant be scope=module, it needs to be unique per test
@pytest.fixture
def setup_deployment():
    deployment_name = gen("app", length=6)
    print(f"\nRunning for deployment {deployment_name}...")
    fbnconfig.deployex(Deployment(deployment_name, resources(deployment_name)), lusid_env, token)
    yield SimpleNamespace(name=deployment_name)
    try:
        print("\nTearing down resources...")
        fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    except Exception:
        pass  # Ignore cleanup errors


@pytest.fixture
def deployment(setup_deployment):
    deployment_name = setup_deployment.name
    return Deployment(deployment_name, resources(deployment_name))


def test_teardown(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_applications_by_client_id(client, f"inttest-app-{deployment_name}")
    assert len(matches) == 0


def test_create(setup_deployment, deployment):
    fbnconfig.deployex(deployment, lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_applications_by_client_id(client, f"inttest-app-{setup_deployment.name}")
    assert len(matches) == 1
    app = matches[0]
    assert app["displayName"] == f"Application{setup_deployment.name}"
    assert app["type"] == "Native"


def test_update(setup_deployment, deployment):
    fbnconfig.deployex(deployment, lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_applications_by_client_id(client, f"inttest-app-{setup_deployment.name}")
    assert len(matches) == 1
    app = matches[0]
    assert app["displayName"] == f"Application{setup_deployment.name}"
    assert app["type"] == "Native"


def get_applications_by_client_id(client, client_id):
    get = client.request("get", "/identity/api/applications")
    get.raise_for_status()
    applications = get.json()
    return [app for app in applications if app["clientId"] == client_id]


@pytest.mark.asyncio
async def test_application_list(setup_deployment, deployment):
    # GIVEN Application is deployed
    fbnconfig.deployex(deployment, lusid_env, token)
    deployment_name = setup_deployment.name
    client_id = f"inttest-app-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list method for that resource
    result = await identity.ApplicationResource.list(
        client=client,
        query_params={},
        path_params={}
    )
    application = next(
        (i for i in result["values"] if i["value"]["clientId"] == client_id)
    )
    application_id = application["value"]["applicationId"]
    application_issuer = application["value"]["issuer"]
    # THEN the result has correct structure
    expected_application = {
        "keys": {
            "applicationId": application_id,
        },
        "value": {
            "id": f"application_{application_id}".lower(),
            "applicationId": application_id,
            "clientId": client_id,
            "type": "Native",
            "displayName": f"Application{deployment_name}",
            "issuer": application_issuer,
            "secret": None
        }
    }
    assert application == expected_application


@pytest.mark.asyncio
async def test_application_fetch(setup_deployment, deployment):
    # GIVEN Application is deployed
    fbnconfig.deployex(deployment, lusid_env, token)
    deployment_name = setup_deployment.name
    client_id = f"inttest-app-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    applications_response = await client.get("/identity/api/applications")
    applications = applications_response.json()
    application = next(
        (app for app in applications if app["clientId"] == client_id)
    )
    application_id = application["id"]
    application_issuer = application["issuer"]
    # WHEN the fetch method is called for that resource
    response = await identity.ApplicationResource.fetch(
        client=client,
        keys={
            "applicationId": application_id,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"application_{application_id}".lower(),
        "applicationId": application_id,
        "clientId": client_id,
        "type": "Native",
        "displayName": f"Application{deployment_name}",
        "issuer": application_issuer,
        "secret": None
    }
    assert response == expected_response



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import customentity, property

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeEntityTypeResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):
        respx_mock.post("/api/api/customentities/entitytypes").mock(
            return_value=httpx.Response(200, json={"entityType": "~typename"})
        )
        # given a desired definition with one field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        # then the state the typename returned by the create call
        assert state == {"entitytype": "~typename"}
        # and a create request was sent without the startValue
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/entitytypes"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "description": "",
                    "collectionType": "Single",
                    "required": False,
                }
            ],
        }

    def test_update_with_no_changes(self, respx_mock):
        # given an existing CE where the field has a description
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "legs",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "",
                            "collectionType": "Single",
                            "required": False,
                        }
                    ],
                },
            )
        )
        # and a desired which is the same but the field description is none
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and a read was made but no PUT

    def test_update_with_collection_single(self, respx_mock):
        # given an existing CE with a collection type of single
        # note, api will not return single here even if the user has
        # set it
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "required": False,
                            # no collectionType member from GET
                        }
                    ],
                },
            )
        )
        # and a desired where the user has explicitly asked for single
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="arms",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    collection_type=customentity.CollectionType.SINGLE,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state None because there is no change
        assert state is None

    def test_update_with_changed_field(self, respx_mock):
        # given an existing CE with an arms field
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        }
                    ],
                },
            )
        )
        respx_mock.put("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(200, json={"entityType": "~whatevah"})
        )
        # and a desired with a legs field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is the same
        assert state == {"entitytype": "~whatevah"}
        # and a put request was sent with legs
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/customentities/entitytypes/~whatevah"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                }
            ],
        }

    def test_update_with_removed_field(self, respx_mock):
        # given an existing CE with arms and legs
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "legs",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        },
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        },
                    ],
                },
            )
        )
        respx_mock.put("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(200, json={"entityType": "~whatevah"})
        )
        # and a desired with a legs field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                )
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is the same
        assert state == {"entitytype": "~whatevah"}
        # and a put request was sent with legs
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/customentities/entitytypes/~whatevah"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                }
            ],
        }

    def test_update_with_additional_field(self, respx_mock):
        # given an existing CE with an arms field
        respx_mock.get("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(
                200,
                json={
                    "entityTypeName": "animal",
                    "displayName": "Animal",
                    "description": "Not mineral or vegetable",
                    "fieldSchema": [
                        {
                            "name": "arms",
                            "lifetime": "Perpetual",
                            "type": "Decimal",
                            "description": "a default descriptoin",
                        }
                    ],
                },
            )
        )
        respx_mock.put("/api/api/customentities/entitytypes/~whatevah").mock(
            return_value=httpx.Response(200, json={"entityType": "~whatevah"})
        )
        # and a desired with an arms and a legs field
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[
                customentity.FieldDefinition(
                    name="legs",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                ),
                customentity.FieldDefinition(
                    name="arms",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.DECIMAL,
                    required=False,
                ),
            ],
        )
        old_state = SimpleNamespace(entitytype="~whatevah")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is the same
        assert state == {"entitytype": "~whatevah"}
        # and a put request was sent with the new fields
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/api/api/customentities/entitytypes/~whatevah"
        assert json.loads(request.content) == {
            "entityTypeName": "animal",
            "displayName": "Animal",
            "description": "Not mineral or vegetable",
            "fieldSchema": [
                {
                    "name": "legs",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                },
                {
                    "name": "arms",
                    "lifetime": "Perpetual",
                    "type": "Decimal",
                    "required": False,
                    "description": "",
                    "collectionType": "Single",
                },
            ],
        }

    def test_delete_throws(self):
        # given a resource that exists in the remnte
        old_state = SimpleNamespace(entitytype="~whatever")
        # when we delete it throws brcause uou cant delete a CE
        with pytest.raises(RuntimeError):
            customentity.EntityTypeResource.delete(self.client, old_state)

    def test_deps(self):
        sut = customentity.EntityTypeResource(
            id="xyz",
            entity_type_name="animal",
            display_name="Animal",
            description="Not mineral or vegetable",
            field_schema=[],
        )
        # it's deps are empty
        assert sut.deps() == []

    def test_dump(self):
        # given an entity type resource
        sut = customentity.EntityTypeResource(
            id="et1",
            entity_type_name="TestEntity",
            display_name="Test Entity Type",
            description="A test entity type",
            field_schema=[
                customentity.FieldDefinition(
                    name="field1",
                    lifetime=customentity.LifeTime.PERPETUAL,
                    type=customentity.FieldType.STRING,
                    collection_type=customentity.CollectionType.SINGLE,
                    required=True,
                    description="First field"
                ),
                customentity.FieldDefinition(
                    name="field2",
                    lifetime=customentity.LifeTime.TIMEVARIANT,
                    type=customentity.FieldType.DECIMAL,
                    collection_type=customentity.CollectionType.ARRAY,
                    required=False,
                    description="Second field"
                )
            ]
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped state is correct
        assert dumped == {
            "entityTypeName": "TestEntity",
            "displayName": "Test Entity Type",
            "description": "A test entity type",
            "fieldSchema": [
                {
                    "name": "field1",
                    "lifetime": "Perpetual",
                    "type": "String",
                    "collectionType": "Single",
                    "required": True,
                    "description": "First field"
                },
                {
                    "name": "field2",
                    "lifetime": "TimeVariant",
                    "type": "Decimal",
                    "collectionType": "Array",
                    "required": False,
                    "description": "Second field"
                }
            ]
        }

    def test_undump(self):
        # given a dumped entity type state
        dumped = {
            "entityTypeName": "TestEntity",
            "displayName": "Test Entity Type",
            "description": "A test entity type",
            "fieldSchema": [
                {
                    "name": "field1",
                    "lifetime": "Perpetual",
                    "type": "String",
                    "collectionType": "Single",
                    "required": True,
                    "description": "First field"
                },
                {
                    "name": "field2",
                    "lifetime": "TimeVariant",
                    "type": "Decimal",
                    "collectionType": "Array",
                    "required": False,
                    "description": "Second field"
                }
            ]
        }
        # when we undump it
        sut = customentity.EntityTypeResource.model_validate(
            dumped,
            context={
                "style": "undump",
                "$refs": {},
                "id": "et1",
            }
        )
        # then the id has been extracted from the context
        assert sut.id == "et1"
        assert sut.entity_type_name == "TestEntity"
        assert sut.display_name == "Test Entity Type"
        assert sut.description == "A test entity type"
        # and the field schema is correctly reconstructed
        assert len(sut.field_schema) == 2
        # first field
        assert sut.field_schema[0].name == "field1"
        assert sut.field_schema[0].lifetime == customentity.LifeTime.PERPETUAL
        assert sut.field_schema[0].type == customentity.FieldType.STRING
        assert sut.field_schema[0].collection_type == customentity.CollectionType.SINGLE
        assert sut.field_schema[0].required is True
        assert sut.field_schema[0].description == "First field"
        # second field
        assert sut.field_schema[1].name == "field2"
        assert sut.field_schema[1].lifetime == customentity.LifeTime.TIMEVARIANT
        assert sut.field_schema[1].type == customentity.FieldType.DECIMAL
        assert sut.field_schema[1].collection_type == customentity.CollectionType.ARRAY
        assert sut.field_schema[1].required is False
        assert sut.field_schema[1].description == "Second field"

    def test_from_api(self, testdata_loader):
        input_data = testdata_loader(
            "entity_type_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "entity_type_resource",
            "list_expected_result"
        )["values"][0]["value"]
        result_data = customentity.EntityTypeResource.from_api(input_data)
        assert result_data == expected_result
        entity_obj = customentity.EntityTypeResource.model_validate(result_data)
        assert entity_obj.id == "entity-type_~cebrsaclhcsftflwfd"

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        input_data = testdata_loader(
            "entity_type_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "entity_type_resource",
            "list_expected_result"
        )
        respx_mock.get(customentity.EntityTypeResource.list_endpoint()).mock(
                    return_value=httpx.Response(200, json=input_data)
                )
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await customentity.EntityTypeResource.list(client, query_params={})
            assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        input_data = testdata_loader(
            "entity_type_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "entity_type_resource",
            "list_expected_result"
        )["values"][0]["value"]
        keys = {"entityType": "type"}
        respx_mock.get(customentity.EntityTypeResource.fetch_endpoint().format(**keys)).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await customentity.EntityTypeResource.fetch(client, keys=keys)
        assert result == expected_result


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeEntityTypeRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_success(self, respx_mock):
        respx_mock.get("/api/api/customentities/entitytypes/TestEntity").mock(
            return_value=httpx.Response(200, json={
                "entityTypeName": "TestEntity",
                "displayName": "Test Entity Type",
                "description": "A test entity type",
                "fieldSchema": []
            })
        )
        # given an entity type ref
        sut = customentity.EntityTypeRef(
            id="et1",
            entity_type_name="TestEntity"
        )
        # when we attach it (should not raise)
        sut.attach(self.client)
        # then a GET request was made
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == "/api/api/customentities/entitytypes/TestEntity"

    def test_attach_not_found(self, respx_mock):
        respx_mock.get("/api/api/customentities/entitytypes/NonExistentEntity").mock(
            return_value=httpx.Response(404, json={"error": "Not found"})
        )
        # given an entity type ref for a non-existent entity
        sut = customentity.EntityTypeRef(
            id="et1",
            entity_type_name="NonExistentEntity"
        )
        # when we try to attach it
        with pytest.raises(RuntimeError) as exc_info:
            sut.attach(self.client)
        # then it raises a descriptive error
        assert "Custom Entity Definition NonExistentEntity does not exist" in str(exc_info.value)

    def test_attach_server_error(self, respx_mock):
        respx_mock.get("/api/api/customentities/entitytypes/TestEntity").mock(
            return_value=httpx.Response(500, json={"error": "Internal server error"})
        )
        # given an entity type ref
        sut = customentity.EntityTypeRef(
            id="et1",
            entity_type_name="TestEntity"
        )
        # when we try to attach it and the server returns an error
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(self.client)

    def test_dump(self):
        # given an entity type ref
        sut = customentity.EntityTypeRef(
            id="et1",
            entity_type_name="TestEntity"
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped state excludes the id field
        assert dumped == {
            "entityTypeName": "TestEntity"
        }

    def test_undump(self):
        # given dumped entity type ref data with id included
        dumped = {
            "entityTypeName": "TestEntity"
        }
        # when we create it directly (since Ref classes don't typically use model_validate with context)
        sut = customentity.EntityTypeRef(
            id="et1",
            entity_type_name=dumped["entityTypeName"]
        )
        # then the fields are correctly set
        assert sut.id == "et1"
        assert sut.entity_type_name == "TestEntity"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeEntityResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture()
    def entity_type(self):
        return customentity.EntityTypeRef(
            id="et1",
            entity_type_name="TestEntity"
        )

    @pytest.fixture()
    def entity_identifier(self):
        return property.DefinitionRef(
            id="id1",
            domain=property.Domain.CustomEntity,
            scope="default",
            code="ClientInternal"
        )

    def test_create(self, respx_mock, entity_type, entity_identifier):
        respx_mock.post("/api/api/customentities/~TestEntity").mock(
            return_value=httpx.Response(200, json={})
        )
        # and a desired entity with fields and identifiers
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Test Entity",
            description="A test entity",
            fields=[
                customentity.EntityField(
                    name="field1",
                    value="test_value"
                )
            ],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="TEST001"
                )
            ]
        )
        # when we create it
        state = sut.create(self.client)
        # then the state contains the entity details
        assert "content_hash" in state
        assert state["entity_type"] == "TestEntity"
        assert state["identifier_type"] == "ClientInternal"
        assert state["identifier_value"] == "TEST001"
        assert state["identifier_scope"] == "default"
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~TestEntity"
        request_data = json.loads(request.content)
        assert request_data == {
            "fields": [{
                "name": "field1", "value": "test_value"
            }],
            "identifiers": [{
                "identifierScope": "default",
                "identifierType": "ClientInternal",
                "identifierValue": "TEST001"
            }],
            "description": "A test entity",
            "displayName": "Test Entity"
        }

    def test_update_no_changes(self, entity_type, entity_identifier):
        # given an entity with a specific content hash
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Test Entity",
            description="A test entity",
            fields=[
                customentity.EntityField(
                    name="field1",
                    value="test_value"
                )
            ],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="TEST001"
                )
            ]
        )
        # and an old state with the same content hash
        desired = sut.model_dump(mode="json", exclude_none=True,
                                 by_alias=True, exclude={"id", "scope", "entity_type"})
        sorted_desired = json.dumps(desired, sort_keys=True)
        from hashlib import sha256
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        old_state = SimpleNamespace(
            entity_type="TestEntity",
            identifier_scope="default",
            identifier_type="ClientInternal",
            identifier_value="TEST001",
            content_hash=content_hash
        )
        # when we update it
        state = sut.update(self.client, old_state)
        # then no update is needed
        assert state is None

    def test_update_with_identifier_changes(self, respx_mock, entity_type, entity_identifier):
        respx_mock.delete("/api/api/customentities/~TestEntity/ClientInternal/TEST001").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.post("/api/api/customentities/~TestEntity").mock(
            return_value=httpx.Response(200, json={})
        )
        # given an entity
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Updated Entity",
            description="An updated entity",
            fields=[
                customentity.EntityField(
                    name="field1",
                    value="updated_value"
                )
            ],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="CHANGED"
                )
            ]
        )
        # and an old state with a different identifier value
        old_state = SimpleNamespace(
            entity_type="TestEntity",
            identifier_scope="default",
            identifier_type="ClientInternal",
            identifier_value="TEST001",
            content_hash="different_hash"
        )
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is updated
        assert state is not None
        assert state["entity_type"] == "TestEntity"
        # and the delete and post calls were made because this is a delete/recrete operation

    def test_update_with_non_identifier_changes(self, respx_mock, entity_type, entity_identifier):
        respx_mock.post("/api/api/customentities/~TestEntity").mock(
            return_value=httpx.Response(200, json={})
        )
        # given an entity
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Updated Entity",
            description="An updated entity",
            fields=[
                customentity.EntityField(
                    name="field1",
                    value="updated_value"
                )
            ],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="TEST001"
                )
            ]
        )
        # and an old state the same identifier but a different hash
        old_state = SimpleNamespace(
            entity_type="TestEntity",
            identifier_scope="default",
            identifier_type="ClientInternal",
            identifier_value="TEST001",
            content_hash="different_hash"
        )
        # when we update it
        sut.update(self.client, old_state)
        # then only a put call is made

    def test_delete(self, respx_mock):
        respx_mock.delete("/api/api/customentities/~TestEntity/ClientInternal/TEST001").mock(
            return_value=httpx.Response(200, json={})
        )
        # given an old state
        old_state = SimpleNamespace(
            entity_type="TestEntity",
            identifier_scope="default",
            identifier_type="ClientInternal",
            identifier_value="TEST001"
        )
        # when we delete it
        customentity.EntityResource.delete(self.client, old_state)
        # then a DELETE request was sent
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/api/api/customentities/~TestEntity/ClientInternal/TEST001"
        assert "identifierScope=default" in str(request.url)

    def test_deps(self, entity_type, entity_identifier):
        # given an entity with an entity type reference
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Test Entity",
            description="A test entity",
            fields=[],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="TEST001"
                )
            ]
        )
        # when we get its dependencies
        deps = sut.deps()
        # then it depends on the entity type
        assert deps == [entity_type, entity_identifier]

    def test_deps_including_property(self, entity_type, entity_identifier):
        # given an entity with a property
        prop_ref = property.DefinitionRef(
            id="p1",
            domain=property.Domain.CustomEntity,
            scope="default",
            code="legCount"
        )
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Test Entity",
            description="A test entity",
            fields=[],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="TEST001"
                )
            ],
            properties=[
                customentity.PropertyValue(
                    property_key=prop_ref,
                    metric_value=customentity.MetricValue(value=3)
                )
            ]
        )
        # when we get its dependencies
        deps = sut.deps()
        # then it depends on the entity type
        assert deps == [entity_type, entity_identifier, prop_ref]

    def test_dump(self, entity_type, entity_identifier):
        # given an entity resource
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Test Entity",
            description="A test entity",
            fields=[
                customentity.EntityField(
                    name="field1",
                    value="test_value"
                ),
                customentity.EntityField(
                    name="field2",
                    value=123.45
                )
            ],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="TEST001",
                )
            ]
        )
        # when we dump it
        dumped = sut.model_dump(
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then the dumped state excludes the id field and uses $ref for references
        assert dumped == {
            "entityType": {
                "$ref": "et1"
            },
            "displayName": "Test Entity",
            "description": "A test entity",
            "fields": [{
                "name": "field1",
                "value": "test_value"
            }, {
                "name": "field2",
                "value": 123.45
            }],
            "identifiers": [{
                "identifierType": {
                    "$ref": "id1"
                },
                "identifierValue": "TEST001",
            }]
        }

    def test_undump(self, entity_type, entity_identifier):
        # given dumped entity data with $ref style references
        dumped = {
            "entityType": {
                "$ref": "et1"
            },
            "displayName": "Test Entity",
            "description": "A test entity",
            "fields": [{
                "name": "field1",
                "value": "test_value"
            }, {
                "name": "field2",
                "value": 123.45
            }],
            "identifiers": [{
                "identifierType": {
                    "$ref": "id1"
                },
                "identifierValue": "TEST001",
            }]
        }
        refs = {
            "et1": entity_type,
            "id1": entity_identifier
        }
        # when we undump it
        sut = customentity.EntityResource.model_validate(
            dumped,
            context={
                "style": "undump",
                "$refs": refs,
                "id": "ent1",
            }
        )
        # then the id has been extracted from the context
        assert sut.id == "ent1"
        assert sut.display_name == "Test Entity"
        assert sut.description == "A test entity"
        # and the entity type is correctly reconstructed
        assert isinstance(sut.entity_type, customentity.EntityTypeRef)
        assert sut.entity_type.entity_type_name == "TestEntity"
        # and the fields are correctly reconstructed
        assert len(sut.fields) == 2
        assert sut.fields[0].name == "field1"
        assert sut.fields[0].value == "test_value"
        assert sut.fields[1].name == "field2"
        assert sut.fields[1].value == 123.45
        # and the identifiers are correctly reconstructed
        assert len(sut.identifiers) == 1
        assert sut.identifiers[0].identifier_value == "TEST001"
        assert sut.identifiers[0].identifier_type == entity_identifier

    def test_read(self, entity_type, entity_identifier, respx_mock):
        # given an entity instance
        respx_mock.get("/api/api/customentities/~TestEntity/ClientInternal/TEST001?identifierScope=default").mock(
            return_value=httpx.Response(
                200,
                json={
                    "href": "https://myco.lusid.com/api/api/customentities/~SupportTicket",
                    "entityType": "~SupportTicket",
                    "version": {
                      "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                      "asAtDate": "2022-03-02T09:00:00.0000000+00:00"
                    },
                    "displayName": "Portfolio Access Denied",
                    "description": "User cannot access the portfolio",
                    "identifiers": [
                      {
                        "identifierScope": "default",
                        "identifierType": "CLientInternal",
                        "identifierValue": "TEST001",
                        "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                        "effectiveUntil": "9999-12-31T23:59:59.9999999+00:00"
                      }
                    ],
                    "fields": [
                      {
                        "name": "field1",
                        "value": "AcmeLtd",
                        "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                        "effectiveUntil": "9999-12-31T23:59:59.9999999+00:00"
                      },
                    ],
                    "relationships": []
                }
            )
        )
        # when we create a desired entity instance
        sut = customentity.EntityResource(
            id="ent1",
            entity_type=entity_type,
            display_name="Test Entity",
            description="A test entity",
            fields=[
                customentity.EntityField(
                    name="field1",
                    value="test_value"
                )
            ],
            identifiers=[
                customentity.EntityIdentifier(
                    identifier_type=entity_identifier,
                    identifier_value="NOT_TEST001"
                )
            ]
        )
        # and we load it from some old state
        old_state = SimpleNamespace(
            entity_type="TestEntity",
            identifier_scope="default",
            identifier_type="ClientInternal",
            identifier_value="TEST001",
            content_hash=None
        )
        result = sut.read(self.client, old_state)
        # then the get request matches the mock
        # and we get a result
        assert result is not None
        assert len(result["fields"]) == 1

    def test_from_api(self, testdata_loader):
        # GIVEN an Entity resource
        input_data = testdata_loader(
            "entity_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "entity_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # WHEN the from_api method is called for that resource
        result = customentity.EntityResource.from_api(input_data)
        # THEN the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        # GIVEN an Entity resource
        input_data = testdata_loader(
            "entity_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "entity_resource",
            "list_expected_result"
        )
        # AND the resource exists in LUSID
        path_params = {
            "entityType": "SupportTicket"
        }
        url = customentity.EntityResource.list_endpoint().format(**path_params)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the list method is called for that resource
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await customentity.EntityResource.list(client, path_params=path_params)
        # THEN the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        # GIVEN an Entity resource
        input_data = testdata_loader(
            "entity_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "entity_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # AND the resource exists in LUSID
        keys = {
            "identifierType": "identifierType",
            "identifierValue": "identifierValue",
            "identifierScope": "identifierScope",
            "entityType": input_data["entityType"]
        }
        url = customentity.EntityResource.fetch_endpoint().format(**keys)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the fetch method is called for that resource
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await customentity.EntityResource.fetch(client, keys=keys)
        # THEN the response is as expected
        assert result == expected_result
        # AND the correct request was made to LUSID
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == url
        assert request.url.query == b"identifierScope=identifierScope"



"""Pydantic models and helpers for the POST /deploy/stream endpoint.

This module contains:
- Request validation models (DeployStreamRequest)
- JSONL event envelope construction (make_event, to_jsonl_line)
- Token redaction (_redact / redact_event)
"""

import json
from datetime import datetime, timezone
from typing import Any

from pydantic import BaseModel, Field, field_validator

# ---------------------------------------------------------------------------
# Request models
# ---------------------------------------------------------------------------


class DeploymentPayload(BaseModel):
    """The deployment plan sent by the caller.

    Must contain a deploymentId (string) and a resources array.
    Each resource is validated later by fbnconfig's own model_validate,
    so here we just check the outer shape.
    """
    deployment_id: str = Field(..., alias="deploymentId", min_length=1)
    resources: list[dict[str, Any]]

    model_config = {"populate_by_name": True}


class DeployStreamRequest(BaseModel):
    """Top-level request body for POST /deploy/stream."""
    deployment: DeploymentPayload

    @field_validator("deployment", mode="before")
    @classmethod
    def deployment_must_be_dict(cls, v):
        if not isinstance(v, dict):
            raise ValueError("deployment must be an object")
        return v


# ---------------------------------------------------------------------------
# JSONL event helpers
# ---------------------------------------------------------------------------

def make_event(
    *,
    deployment_id: str,
    request_id: str,
    correlation_id: str | None = None,
    event_type: str,
    status: str,
    sequence: int,
    data: dict[str, Any] | None = None,
) -> dict[str, Any]:
    """Build a JSONL event envelope.

    Every line in the stream shares this shape - the 'data' field carries
    event-specific payload.
    """
    event = {
        "ts": datetime.now(timezone.utc).isoformat(),
        "deploymentId": deployment_id,
        "requestId": request_id,
        "eventType": event_type,
        "status": status,
        "sequence": sequence,
        "data": data or {},
    }
    if correlation_id is not None:
        event["correlationId"] = correlation_id
    return event


def to_jsonl_line(event: dict[str, Any]) -> str:
    """Serialise an event dict to a single JSONL line (with trailing newline)."""
    return json.dumps(event, separators=(",", ":")) + "\n"


def _redact_value(value: Any, secrets: set[str]) -> Any:
    """Recursively replace any occurrence of a known secret with '********'."""
    if isinstance(value, str):
        for secret in secrets:
            if secret in value:
                value = value.replace(secret, "********")
        return value
    if isinstance(value, dict):
        return {k: _redact_value(v, secrets) for k, v in value.items()}
    if isinstance(value, list):
        return [_redact_value(item, secrets) for item in value]
    return value


def redact_event(event: dict[str, Any], secrets: set[str]) -> dict[str, Any]:
    """Return a copy of *event* with any known secret strings scrubbed."""
    return _redact_value(event, secrets)



from typing import ClassVar

from pydantic import BaseModel, Field

from fbnconfig import schemagen
from fbnconfig.resource_abc import CamelAlias, Ref, Resource


class SimpleModel(BaseModel, CamelAlias):
    id: str
    x_value: int
    y_value: int


class TestRef(BaseModel, Ref, CamelAlias):
    id: str
    # this is what the register decorator puts on
    resource_type: ClassVar = "TestRef"
    no_init: str = Field(None, init=False)
    normal_field: str


class TestResource(BaseModel, Resource, CamelAlias):
    id: str
    resource_type: ClassVar = "TestRef"
    simple: SimpleModel
    optional: int | None = None
    unioned: int | str | None = None


def test_simple_model():
    # given a simple (not resource or ref) model, when we generate the schema
    schema = schemagen.schema(SimpleModel)
    assert schema["properties"]
    # it contains the fields in camelCase
    assert schema["properties"]["xValue"]
    assert schema["properties"]["yValue"]
    # it keeps it's id field because it's not a resource
    assert schema["properties"]["id"]
    # the camel fields get converted to a title case title
    assert schema["properties"]["xValue"]["title"] == "X Value"


def test_ref():
    # given a ref class when we generate the schema
    schema = schemagen.schema(TestRef)
    assert schema["properties"]
    # it contains the fields in camelCase
    assert schema["properties"]["normalField"]
    # the id field is removed
    assert schema["properties"].get("id", None) is None
    # the non-init field is marked as readonly
    assert schema["properties"]["noInit"]["readOnly"] is True


def test_resource():
    # given a ref class when we generate the schema
    schema = schemagen.schema(TestResource)
    # the embedded type will get put into $defs
    assert schema["$defs"].get("SimpleModel", None) is not None
    # the primary type will go into the top
    assert schema["properties"]
    assert schema["properties"]["optional"]
    # the anyOf which pydantic puts in for the nullable field is simplified
    assert schema["properties"]["optional"]["type"] == "integer"
    # genuine unions are preserved
    assert schema["properties"]["unioned"]
    assert schema["properties"]["unioned"]["anyOf"]



from fbnconfig import Deployment, horizon


def configure(_):
    instance = horizon.IntegrationInstanceResource(
        id="instance",
        integration_type="blackrock-aladdin",
        name="example_aladdin_Int96",
        description="my first aladdin",
        enabled=False,
        triggers=[],
        details={
            "sourceFileLocation": "/sftp",
            "onboardingDate": "2022-01-01",
            "localTimeZone": "Australia/Melbourne",
            "unpackFromArchive": False,
            "archiveFileMask": "",
            "isFxNdfAllowed": True,
            "int29AnalyticsFileType": "NotApplicable",
            "allowCoreFieldUpdates": "Never",
            "blackRockAladdinInterfaceSelections": [{
                "interfaceNumber": "Int96",
                "fileMask": "^ATOM_R_newcash96_daily.*_*.xml$",
                "localCutTime": "00:00"
            }]

        }
    )
    return Deployment("example-aladdin", [instance])



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
import tests.integration.datatype as example
from fbnconfig import datatype
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture
def setup_deployment():
    deployment_name = gen("datatype")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    fbnconfig.deployex(example.configure(setup_deployment), lusid_env, token)

    fbnconfig.deployex(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    try:
        client.get(f"/api/api/datatypes/{setup_deployment.name}/robtest_strategy")
    except HTTPStatusError as error:
        # check it was deleted
        assert error.response.status_code == 404


def test_create(setup_deployment):
    fbnconfig.deployex(example.configure(setup_deployment), lusid_env, token)
    res = client.get(f"/api/api/datatypes/{setup_deployment.name}/robtest_priority").json()
    assert res is not None


def test_nochange(setup_deployment):
    fbnconfig.deployex(example.configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deployex(example.configure(setup_deployment), lusid_env, token)
    print(update)
    assert [a.change for a in update] == ["nochange"] * 3


@pytest.mark.asyncio
async def test_list(setup_deployment):
    # GIVEN DataType is deployed
    fbnconfig.deployex(example.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    code = "robtest_priority"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for that code
    result = await datatype.DataTypeResource.list(
        client=client,
        query_params={
            "filter": f"id.code eq '{code}' and id.scope eq '{scope}'"
        }
    )
    # THEN the result matches expected structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "code": code,
                    "scope": scope,
                },
                "value": {
                    "id": f"datatype_{scope}_{code}".lower(),
                    "scope": scope,
                    "code": code,
                    "acceptableUnits": [],
                    "acceptableValues": [
                        "High",
                        "Low",
                        "Medium",
                    ],
                    "description": "A test datatype for Priority",
                    "displayName": "Priority Test",
                    "typeValueRange": "Closed",
                    "unitSchema": "NoUnits",
                    "valueType": "String",
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_fetch(setup_deployment):
    # GIVEN DataType is deployed
    fbnconfig.deployex(example.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    code = "robtest_priority"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await datatype.DataTypeResource.fetch(
        client=client,
        keys={
            "scope": scope,
            "code": code
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"datatype_{scope}_{code}".lower(),
        "scope": scope,
        "code": code,
        "acceptableUnits": [],
        "acceptableValues": [
            "High",
            "Low",
            "Medium"
        ],
        "description": "A test datatype for Priority",
        "displayName": "Priority Test",
        "valueType": "String",
        "typeValueRange": "Closed",
        "unitSchema": "NoUnits",
        "referenceData": {
            "fieldDefinitions": [],
            "values": []
        },
    }
    assert response == expected_response



import os
from types import SimpleNamespace
from unittest.mock import ANY

import pytest
from pytest import fixture

import fbnconfig
from fbnconfig import identity, scheduler
from fbnconfig import workflows as wf
from tests.integration.generate_test_name import gen


@fixture(scope="module")
def lusid_env():
    if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
        raise (
            RuntimeError("FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
        )

    env = os.environ["LUSID_ENV"]
    token = os.environ["FBN_ACCESS_TOKEN"]
    return SimpleNamespace(env=env, token=token)


@fixture(scope="module")
def client(lusid_env):
    return fbnconfig.create_client(lusid_env.env, lusid_env.token)


@fixture(scope="module")
def deployment_name():
    return gen("workflows")


def resources(deployment_name):
    tag = "v5.0"
    image = scheduler.ImageResource(
        id="img1",
        source_image="harbor.finbourne.com/ceng/fbnconfig-pipeline:0.1",
        dest_name="beany",
        dest_tag=tag,
    )
    job = scheduler.JobResource(
        id="job1",
        scope=deployment_name,
        code="jobwf",
        image=image,
        name="wfjob",
        description="something nice",
        min_cpu="1",
        max_cpu="2",
        argument_definitions={},
    )
    worker = wf.WorkerResource(
        id="wr3-example",
        scope=deployment_name,
        code="wr3",
        display_name="I am worker3",
        worker_configuration=wf.SchedulerJob(job=job),
    )
    pending_state = wf.TaskStateDefinition(name="pending")
    in_progress_state = wf.TaskStateDefinition(name="inprogress")
    end_state = wf.TaskStateDefinition(name="end")
    start_trigger = wf.TriggerDefinition(name="start", type="External")
    end_trigger = wf.TriggerDefinition(name="end", type="External")
    some_field = wf.TaskFieldDefinition(name="imafield", type=wf.TaskFieldDefinitionType.STRING)
    do_something_action = wf.ActionDefinition(
        name="start-something-worker",
        action_details=wf.RunWorkerAction(
            worker=worker,
            worker_parameters={},
            worker_status_triggers=wf.WorkerStatusTriggers(completed_with_results=end_trigger),
        ),
    )
    task_def = wf.TaskDefinitionResource(
        id="integrationtest-task-definition",
        scope=deployment_name,
        code="DoSomething",
        display_name="Does something",
        description="Task description",
        states=[pending_state, in_progress_state, end_state],
        field_schema=[some_field],
        initial_state=wf.InitialState(name=pending_state),
        triggers=[start_trigger, end_trigger],
        transitions=[
            wf.TaskTransitionDefinition(
                from_state=pending_state,
                to_state=in_progress_state,
                trigger=start_trigger,
                action=do_something_action,
            ),
            wf.TaskTransitionDefinition(
                from_state=in_progress_state, to_state=end_state, trigger=end_trigger
            ),
        ],
        actions=[do_something_action],
    )
    event_handler = wf.EventHandlerResource(
        id="integrationtest-eventhandler",
        scope=deployment_name,
        code="testeventhandler",
        display_name="new name",
        description="something",
        status=wf.EventStatus.INACTIVE,
        event_matching_pattern=wf.EventMatchingPattern(
            event_type="FileCreated",
            filter="body.filePath startswith '/somedomain/quotes'"),
        run_as_user_id=wf.EventHandlerMapping(
            map_from="header.userId"
        ),
        task_definition=task_def,
        task_activity=wf.CreateNewTaskActivity(
            correlation_ids=[wf.EventHandlerMapping(set_to="int-test")],
            task_fields={
                some_field: wf.FieldMapping(map_from="body.filePath")  # pyright: ignore
            },
            initial_trigger=start_trigger
        )
    )
    current_user = identity.CurrentUserRef(
        id="me-user"
    )
    schedule_handler = wf.EventHandlerResource(
        id="integrationtest-schedule-handler",
        scope=deployment_name,
        code="testschedulehandler",
        display_name="new name",
        description="something",
        status=wf.EventStatus.INACTIVE,
        schedule_matching_pattern=wf.ScheduleMatchingPattern(
            context=wf.ScheduleMatchingPatternContext(
                time_zone="America/New_York",
            ),
            recurrence_pattern=wf.RecurrencePattern(
                time_constraints=wf.TimeConstraints(
                    start_date="2024-01-01",
                    times_of_day=[
                        wf.SpecifiedTime(hours=6, minutes=30)
                    ]
                ),
                date_regularity=wf.WeekRegularity(
                    days_of_week=["Monday"],
                    frequency=2
                ),
                business_day_adjustment="None"
            )
        ),
        run_as_user_id=wf.EventHandlerMapping(
            set_to=current_user
        ),
        task_definition=task_def,
        task_activity=wf.CreateNewTaskActivity(
            correlation_ids=[wf.EventHandlerMapping(set_to="int-test")],
            task_fields={
                some_field: wf.FieldMapping(set_to="aconstant")  # pyright: ignore
            },
            initial_trigger=start_trigger
        )
    )
    return {
        "image": image,
        "job": job,
        "worker": worker,
        "task_def": task_def,
        "event_handler": event_handler,
        "schedule_handler": schedule_handler
    }


@fixture
def deployment(deployment_name, lusid_env):
    res = resources(deployment_name)
    print(f"\nRunning for deployment {deployment_name}...")
    yield fbnconfig.Deployment(deployment_name, list(res.values()))
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env.env, lusid_env.token)


def test_teardown(deployment, client, lusid_env):
    # create first
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    fbnconfig.deployex(fbnconfig.Deployment(deployment.id, []), lusid_env.env, lusid_env.token)
    result = client.get(f"/workflow/api/workers?filter=id.scope eq '{deployment.id}'")
    # no response returned
    assert len(result.json()["values"]) == 0


def test_create(deployment, client, lusid_env):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    workers = client.get(f"/workflow/api/workers?filter=id.scope eq '{deployment.id}'")
    task_defs = client.get(f"/workflow/api/taskdefinitions?filter=id.scope eq '{deployment.id}'")
    event_handlers = client.get(
        f"/workflow/api/eventhandlers?filter=id.scope eq '{deployment.id}'"
    )
    assert len(workers.json()["values"]) == 1
    assert len(task_defs.json()["values"]) == 1
    assert len(event_handlers.json()["values"]) == 2


def test_update_nochange(deployment, client, lusid_env):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    actions = fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    assert {a.change for a in actions} == {"nochange", "attach"}
    workers = client.get(f"/workflow/api/workers?filter=id.scope eq '{deployment.id}'")
    task_defs = client.get(f"/workflow/api/taskdefinitions?filter=id.scope eq '{deployment.id}'")
    event_handlers = client.get(
        f"/workflow/api/eventhandlers?filter=id.scope eq '{deployment.id}'"
    )
    assert len(workers.json()["values"]) == 1
    assert len(task_defs.json()["values"]) == 1
    assert len(event_handlers.json()["values"]) == 2


@pytest.mark.asyncio
async def test_worker_list_with_filter(deployment, lusid_env):
    # GIVEN Worker is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    scope = deployment_name
    code = "wr3"
    job_scope = deployment_name
    job_code = "jobwf"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await wf.WorkerResource.list(
        client=client,
        query_params={"filter": f"id.code eq '{code}' and id.scope eq '{scope}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "code": code,
                    "scope": scope
                },
                "value": {
                    "id": f"worker_{scope}_{code}".lower(),
                    "code": code,
                    "scope": scope,
                    "displayName": "I am worker3",
                    "parameters": [],
                    "resultFields": [],
                    "workerConfiguration": {
                        "type": "SchedulerJob",
                        "job": {
                            "$ref": f"job_{job_scope}_{job_code}".lower()
                        }
                    }
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_worker_fetch(deployment, lusid_env):
    # GIVEN Worker is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    scope = deployment_name
    code = "wr3"
    job_scope = deployment_name
    job_code = "jobwf"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await wf.WorkerResource.fetch(
        client=client,
        keys={
            "code": code,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"worker_{scope}_{code}".lower(),
        "code": code,
        "scope": scope,
        "displayName": "I am worker3",
        "parameters": [],
        "resultFields": [],
        "workerConfiguration": {
            "type": "SchedulerJob",
            "job": {
                "$ref": f"job_{job_scope}_{job_code}".lower()
            }
        }
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_task_definition_list_with_filter(deployment, lusid_env):
    # GIVEN TaskDefinition is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    scope = deployment_name
    code = "DoSomething"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await wf.TaskDefinitionResource.list(
        client=client,
        query_params={"filter": f"id.code eq '{code}' and id.scope eq '{scope}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "code": code,
                    "scope": scope
                },
                "value": {
                    "id": f"task-definition_{scope}_{code}".lower(),
                    "code": code,
                    "scope": scope,
                    "description": "Task description",
                    "displayName": "Does something",
                    "initialState": {
                        "name": "pending",
                        "requiredFields": [],
                    },
                    "fieldSchema": [
                        {
                            "name": "imafield",
                            "type": "String",
                        }
                    ],
                    "actions": [
                        {
                            "name": "start-something-worker",
                            "actionDetails": {
                                "type": "RunWorker",
                                "workerId": {
                                    "code": "wr3",
                                    "scope": scope,
                                },
                                "workerParameters": {},
                                "workerAsAt": ANY,
                                "workerStatusTriggers": {
                                    "completedNoResults": None,
                                    "failedToComplete": None,
                                    "failedToStart": None,
                                    "started": None,
                                    "completedWithResults": "end",
                                },
                                "workerTimeout": None,
                                "childTaskConfigurations": [],
                            }
                        }
                    ],
                    "states": [
                        {
                            "name": "pending",
                        },
                        {
                            "name": "inprogress",
                        },
                        {
                            "name": "end",
                        },
                    ],
                    "transitions": [
                        {
                            "action": "start-something-worker",
                            "fromState": "pending",
                            "toState": "inprogress",
                            "trigger": "start",
                        },
                        {
                            "fromState": "inprogress",
                            "toState": "end",
                            "trigger": "end",
                        }
                    ],
                    "triggers": [
                        {
                            "name": "start",
                            "type": "External",
                        },
                        {
                            "name": "end",
                            "type": "External",
                        }
                    ],
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_task_definition_fetch(deployment, lusid_env):
    # GIVEN TaskDefinition is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    scope = deployment_name
    code = "DoSomething"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await wf.TaskDefinitionResource.fetch(
        client=client,
        keys={
            "code": code,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"task-definition_{scope}_{code}".lower(),
        "code": code,
        "scope": scope,
        "description": "Task description",
        "displayName": "Does something",
        "initialState": {
            "name": "pending",
            "requiredFields": [],
        },
        "fieldSchema": [
            {
                "name": "imafield",
                "type": "String",
            }
        ],
        "actions": [
            {
                "name": "start-something-worker",
                "actionDetails": {
                    "type": "RunWorker",
                    "workerId": {
                        "code": "wr3",
                        "scope": scope,
                    },
                    "workerParameters": {},
                    "workerAsAt": ANY,
                    "workerStatusTriggers": {
                        "completedNoResults": None,
                        "failedToComplete": None,
                        "failedToStart": None,
                        "started": None,
                        "completedWithResults": "end",
                    },
                    "workerTimeout": None,
                    "childTaskConfigurations": [],
                }
            }
        ],
        "states": [
            {
                "name": "pending",
            },
            {
                "name": "inprogress",
            },
            {
                "name": "end",
            },
        ],
        "transitions": [
            {
                "action": "start-something-worker",
                "fromState": "pending",
                "toState": "inprogress",
                "trigger": "start",
            },
            {
                "fromState": "inprogress",
                "toState": "end",
                "trigger": "end",
            }
        ],
        "triggers": [
            {
                "name": "start",
                "type": "External",
            },
            {
                "name": "end",
                "type": "External",
            }
        ],
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_event_handler_list_with_filter(deployment, lusid_env):
    # GIVEN EventHandler is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    scope = deployment_name
    code = "testschedulehandler"
    task_def_code = "DoSomething"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await wf.EventHandlerResource.list(
        client=client,
        query_params={"filter": f"id.code eq '{code}' and id.scope eq '{scope}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "code": code,
                    "scope": scope
                },
                "value": {
                    "id": f"event-handler_{scope}_{code}".lower(),
                    "code": code,
                    "scope": scope,
                    "description": "something",
                    "displayName": "new name",
                    "runAsUserId": {
                        "setTo": ANY
                    },
                    "scheduleMatchingPattern": {
                        "context": {
                            "holidayCalendars": [],
                            "timeZone": "America/New_York",
                        },
                        "recurrencePattern": {
                            "businessDayAdjustment": "None",
                            "dateRegularity": {
                                "daysOfWeek": [
                                    "Monday",
                                ],
                                "type": "Week",
                                "frequency": 2,
                            },
                            "timeConstraints": {
                                "startDate": "2024-01-01",
                                "timesOfDay": [
                                    {
                                        "hours": 6,
                                        "minutes": 30,
                                        "type": "Specified",
                                    }
                                ]
                            },
                        },
                    },
                    "status": "Inactive",
                    "taskDefinitionId": {
                        "$ref": f"task-definition_{scope}_{task_def_code}".lower()
                    },
                    "taskDefinitionAsAt": ANY,
                    "taskActivity": {
                        "correlationIds": [
                            {
                                "setTo": "int-test"
                            }
                        ],
                        "initialTrigger": "start",
                        "taskFields": {
                            "imafield": {
                                "setTo": "aconstant"
                            }
                        },
                        "type": "CreateNewTask",
                        "scheduleDependentTaskFields": {}
                    }
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_event_handler_fetch(deployment, lusid_env):
    # GIVEN EventHandler is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    scope = deployment_name
    code = "testschedulehandler"
    task_def_code = "DoSomething"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await wf.EventHandlerResource.fetch(
        client=client,
        keys={
            "code": code,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"event-handler_{scope}_{code}".lower(),
        "code": code,
        "scope": scope,
        "description": "something",
        "displayName": "new name",
        "runAsUserId": {
            "setTo": ANY
        },
        "scheduleMatchingPattern": {
            "context": {
                "holidayCalendars": [],
                "timeZone": "America/New_York",
            },
            "recurrencePattern": {
                "businessDayAdjustment": "None",
                "dateRegularity": {
                    "daysOfWeek": [
                        "Monday",
                    ],
                    "type": "Week",
                    "frequency": 2,
                },
                "timeConstraints": {
                    "startDate": "2024-01-01",
                    "timesOfDay": [
                        {
                            "hours": 6,
                            "minutes": 30,
                            "type": "Specified",
                        }
                    ]
                },
            },
        },
        "status": "Inactive",
        "taskDefinitionId": {
            "$ref": f"task-definition_{scope}_{task_def_code}".lower()
        },
        "taskDefinitionAsAt": ANY,
        "taskActivity": {
            "correlationIds": [
                {
                    "setTo": "int-test"
                }
            ],
            "initialTrigger": "start",
            "taskFields": {
                "imafield": {
                    "setTo": "aconstant"
                }
            },
            "type": "CreateNewTask",
            "scheduleDependentTaskFields": {}
        }
    }
    assert response == expected_response
